
import netCDF4 as nc
import numpy as np

from smrf.distribute import image_data
from smrf.envphys import precip, Snow, storms
from smrf.utils import utils


class ppt(image_data.image_data):
    """
    The :mod:`~smrf.distribute.precip.ppt` class allows for variable specific
    distributions that go beyond the base class.

    The instantaneous precipitation typically has a positive trend with
    elevation due to orographic effects. However, the precipitation
    distribution can be further complicated for storms that have isolated
    impact at only a few measurement locations, for example thunderstorms
    or small precipitation events. Some distribution methods may be better
    suited than others for capturing the trend of these small events with
    multiple stations that record no precipitation may have a negative impact
    on the distribution.

    The precipitation phase, or the amount of precipitation falling as rain or
    snow, can significantly alter the energy and mass balance of the snowpack,
    either leading to snow accumulation or inducing melt :cite:`Marks&al:1998`
    :cite:`Kormos&al:2014`. The precipitation phase and initial snow density
    estimated using a variety of models that can be set in the configuration
    file.

    For more information on the available models, checkout
    :mod:`~smrf.envphys.snow`.

    After the precipitation phase is calculated, the storm information can be
    determined. The spatial resolution for which storm definitions are applied
    is based on the snow model thats selected.

    The time since last storm is based on an accumulated precipitation mass
    threshold, the time elapsed since it last snowed, and the precipitation
    phase.  These factors determine the start and end time of a storm that
    has produced enough precipitation as snow to change the surface albedo.

    Args:
        pptConfig: The [precip] section of the configuration file
        time_step: The time step in minutes of the data, defaults to 60

    Attributes:
        config: configuration from [precip] section
        precip: numpy array of the precipitation
        percent_snow: numpy array of the percent of time step that was snow
        snow_density: numpy array of the snow density
        storm_days: numpy array of the days since last storm
        storm_total: numpy array of the precipitation mass for the storm
        last_storm_day: numpy array of the day of the last storm (decimal day)
        min: minimum value of precipitation is 0
        max: maximum value of precipitation is infinite
        stations: stations to be used in alphabetical order

    """

    variable = 'precip'

    # these are variables that can be output
    OUTPUT_VARIABLES = {
        'precip': {
            'units': 'mm',
            'standard_name': 'precipitation_mass',
            'long_name': 'Precipitation mass'
        },
        'percent_snow': {
            'units': '%',
            'standard_name': 'percent_snow',
            'long_name': 'Percent of precipitation as snow'
        },
        'snow_density': {
            'units': 'kg/m3',
            'standard_name': 'snow_density',
            'long_name': 'Precipitation snow density'
        },
        'storm_days': {
            'units': 'day',
            'standard_name': 'days_since_last_storm',
            'long_name': 'Days since the last storm'
        },
        'storm_total': {
            'units': 'mm',
            'standard_name': 'precipitation_mass_storm',
            'long_name': 'Precipitation mass for the storm period'
        },
        'last_storm_day': {
            'units': 'day',
            'standard_name': 'day_of_last_storm',
            'long_name': 'Decimal day of the last storm since Oct 1'
        }
    }

    # these are variables that are operate at the end only and do not need to
    # be written during main distribute loop
    post_process_variables = {}

    BASE_THREAD_VARIABLES = frozenset([
        'precip',
        'percent_snow',
        'snow_density',
        'storm_days',
        'storm_total'
    ])

    def __init__(self, pptConfig, start_date, time_step=60):

        # extend the base class
        image_data.image_data.__init__(self, self.variable)

        # check and assign the configuration
        self.getConfig(pptConfig)
        self.time_step = float(time_step)
        self.start_date = start_date

    def initialize(self, topo, data, date_time=None):

        self._logger.debug('Initializing distribute.precip')

        self.date_time = date_time
        self._initialize(topo, data.metadata)
        self.percent_snow = np.zeros((topo.ny, topo.nx))
        self.snow_density = np.zeros((topo.ny, topo.nx))
        self.storm_days = np.zeros((topo.ny, topo.nx))
        self.storm_total = np.zeros((topo.ny, topo.nx))
        self.last_storm_day = np.zeros((topo.ny, topo.nx))
        self.dem = topo.dem

        # Assign storm_days array if given
        if self.config["storm_days_restart"] is not None:
            self._logger.debug('Reading {} from {}'.format(
                'storm_days', self.config['storm_days_restart']))
            f = nc.Dataset(self.config['storm_days_restart'], 'r')
            f.set_always_mask(False)

            if 'storm_days' in f.variables:
                t = f.variables['time']
                time = nc.num2date(
                    t[:],
                    t.getncattr('units'),
                    t.getncattr('calendar'),
                    only_use_cftime_datetimes=False)
                time = np.array(
                    [ti.replace(tzinfo=self.start_date.tzinfo) for ti in time])
                time_ind = np.where(time == self.start_date.to_pydatetime())[0]

                if time_ind.size == 0:
                    self._logger.warning(
                        'Invalid storm_days input! Setting to 0.0')
                    self.storm_days = np.zeros((topo.ny, topo.nx))

                else:
                    # start at index of storm_days - 1
                    self.storm_days = f.variables['storm_days'][time_ind - 1, :, :][0]  # noqa
            else:
                self._logger.error(
                    'Variable storm_days not in {}'.format(
                        self.config['storm_days_restart'])
                )

            f.close()

        self.ppt_threshold = self.config['storm_mass_threshold']

        # Time steps needed to end a storm definition
        self.time_to_end_storm = self.config['marks2017_timesteps_to_end_storms']  # noqa

        self.nasde_model = self.config['new_snow_density_model']

        self._logger.info(
            """Using {0} for the new accumulated snow density"""
            """ model:  """.format(self.nasde_model))

        if self.nasde_model == 'marks2017':
            self.add_thread_variables('storm_id')

            self.storm_total = np.zeros((topo.ny, topo.nx))

            self.storms = []
            self.time_steps_since_precip = 0
            self.storming = False

            # Clip and adjust the precip data so that there is only precip
            # during the storm and ad back in the missing data to conserve mass
            data.precip = data.precip[self.stations]
            self.storms, storm_count = storms.tracking_by_station(
                data.precip,
                mass_thresh=self.ppt_threshold,
                steps_thresh=self.time_to_end_storm)
            self.corrected_precip = storms.clip_and_correct(
                data.precip,
                self.storms,
                stations=self.stations)

            if storm_count != 0:
                self._logger.info(
                    "Identified Storms:\n{0}".format(self.storms))
                self.storm_id = 0
                self._logger.info(
                    "Estimated number of storms: {0}".format(storm_count))

            else:
                if (data.precip.sum() > 0).any():
                    self.storm_id = np.nan
                    self._logger.warning("Zero events triggered a storm "
                                         " definition, None of the precip will"
                                         " be used in this run.")

        # if redistributing due to wind
        if self.config['precip_rescaling_model'] == 'winstral':
            self._tbreak_file = nc.Dataset(
                self.config['winstral_tbreak_netcdf'], 'r')
            self.tbreak = self._tbreak_file.variables['tbreak'][:]
            self.tbreak_direction = self._tbreak_file.variables['direction'][:]
            self._tbreak_file.close()
            self._logger.debug('Read data from {}'
                               .format(self.config['winstral_tbreak_netcdf']))

            # get the veg values
            self.veg_type = topo.veg_type
            matching = [s for s in self.config.keys() if "winstral_veg_" in s]
            v = {}
            for m in matching:
                if m != 'winstral_veg_default':
                    ms = m.split('_')
                    # v[ms[1]] = float(self.config[m])
                    if type(self.config[m]) == list:
                        v[ms[1]] = float(self.config[m][0])
                    else:
                        v[ms[1]] = float(self.config[m])
            self.veg = v

    def distribute(self, data, dpt, precip_temp, ta, time, wind, temp,
                   wind_direction=None, dir_round_cell=None, wind_speed=None,
                   cell_maxus=None):
        """
        Distribute given a Panda's dataframe for a single time step. Calls
        :mod:`smrf.distribute.image_data.image_data._distribute`.

        The following steps are taken when distributing precip, if there is
        precipitation measured:

        1. Distribute the instantaneous precipitation from the measurement data
        2. Determine the distributed precipitation phase based on the
            precipitation temperature
        3. Calculate the storms based on the accumulated mass, time since last
            storm, and precipitation phase threshold


        Args:
            data:           Pandas dataframe for a single time step from precip
            dpt:            dew point numpy array that will be used for
            precip_temp:    numpy array of the precipitation temperature
            ta:             air temp numpy array
            time:           pass in the time were are currently on
            wind:           station wind speed at time step
            temp:           station air temperature at time step
            wind_direction: numpy array for simulated wind direction
            dir_round_cell: numpy array for wind direction in discreet
                            increments for referencing maxus at a specific
                            direction
            wind_speed:     numpy array of wind speed
            cell_maxus:     numpy array for maxus at correct wind directions
        """

        self._logger.debug('%s Distributing all precip' % data.name)
        data = data[self.stations]

        if self.config['distribution'] != 'grid':
            if self.nasde_model == 'marks2017':
                # Adjust the precip for undercatchment
                if self.config['station_adjust_for_undercatch']:
                    self._logger.debug(
                        '%s Adjusting precip for undercatch...' % data.name)
                    self.corrected_precip.loc[time] = \
                        precip.adjust_for_undercatch(
                            self.corrected_precip.loc[time],
                        wind,
                        temp,
                        self.config,
                        self.metadata)

                # Use the clipped and corrected precip
                self.distribute_for_marks2017(
                    self.corrected_precip.loc[time],
                    precip_temp,
                    ta,
                    time)

            else:
                # Adjust the precip for undercatchment
                if self.config['station_adjust_for_undercatch']:
                    self._logger.debug(
                        '%s Adjusting precip for undercatch...' % data.name)
                    data = precip.adjust_for_undercatch(
                        data,
                        wind,
                        temp,
                        self.config,
                        self.metadata)

                    self.distribute_for_susong1999(
                        data, precip_temp, time)
        else:
            self.distribute_for_susong1999(data, precip_temp, time)

        # redistribute due to wind to account for driftin
        if self.config['precip_rescaling_model'] == 'winstral':
            self._logger.debug('%s Redistributing due to wind' % data.name)
            if np.any(dpt < 0.5):
                self.precip = precip.dist_precip_wind(
                    self.precip,
                    dpt,
                    wind_direction,
                    dir_round_cell,
                    wind_speed,
                    cell_maxus,
                    self.tbreak,
                    self.tbreak_direction,
                    self.veg_type,
                    self.veg,
                    self.config)

    def distribute_for_marks2017(self, data, precip_temp, ta, time):
        """
        Specialized distribute function for working with the new accumulated
        snow density model Marks2017 requires storm total and a corrected
        precipitation as to avoid precip between storms.
        """
        # self.corrected_precip # = data.mul(self.storm_correction)

        if data.sum() > 0.0:
            # Check for time in every storm
            for i, s in self.storms.iterrows():
                storm_start = s['start']
                storm_end = s['end']

                if time >= storm_start and time <= storm_end:
                    # establish storm info
                    self.storm_id = i
                    storm = self.storms.iloc[self.storm_id]
                    self.storming = True
                    break
                else:
                    self.storming = False

            self._logger.debug("Storming? {0}".format(self.storming))
            self._logger.debug("Current Storm ID = {0}".format(self.storm_id))

            # distribute data and set the min/max
            self._distribute(data, zeros=None)
            self.precip = utils.set_min_max(self.precip, self.min, self.max)

            if time == storm_start:
                # Entered into a new storm period distribute the storm total
                self._logger.debug('{0} Entering storm #{1}'
                                   .format(data.name, self.storm_id+1))
                if precip_temp.min() < 2.0:
                    self._logger.debug('''Distributing Total Precip
                                        for Storm #{0}'''
                                       .format(self.storm_id+1))
                    self._distribute(storm[self.stations].astype(float),
                                     other_attribute='storm_total')
                    self.storm_total = utils.set_min_max(self.storm_total,
                                                         self.min,
                                                         self.max)

            if self.storming and precip_temp.min() < 2.0:
                self._logger.debug('''Calculating new snow density for
                                    storm #{0}'''.format(self.storm_id+1))
                # determine the precip phase and den
                snow_den, perc_snow = Snow.phase_and_density(
                    precip_temp,
                    self.precip,
                    nasde_model=self.nasde_model)

            else:
                snow_den = np.zeros(self.precip.shape)
                perc_snow = np.zeros(self.precip.shape)

            # calculate decimal days since last storm
            self.storm_days = storms.time_since_storm_pixel(
                self.precip,
                precip_temp,
                perc_snow,
                storming=self.storming,
                time_step=self.time_step/60.0/24.0,
                stormDays=self.storm_days,
                mass=self.ppt_threshold)

        else:
            self.storm_days += self.time_step/60.0/24.0
            self.precip = np.zeros(self.storm_days.shape)
            perc_snow = np.zeros(self.storm_days.shape)
            snow_den = np.zeros(self.storm_days.shape)

        # save the model state
        self.percent_snow = perc_snow
        self.snow_density = snow_den

        # day of last storm, this will be used in albedo
        self.last_storm_day = utils.water_day(data.name)[0] - \
            self.storm_days - 0.001

    def distribute_for_susong1999(self, data, ppt_temp, time):
        """Susong 1999 estimates percent snow and snow density based on
        Susong et al, (1999) :cite:`Susong&al:1999`.

        Args:
            data (pd.DataFrame): Precipitation mass data
            ppt_temp (pd.DataFrame): Precipitation temperature data
            time : Unused
        """

        if data.sum() > 0:

            self._distribute(data)
            self.precip = utils.set_min_max(self.precip, self.min, self.max)

            # determine the precip phase and den
            snow_den, perc_snow = Snow.phase_and_density(
                ppt_temp,
                self.precip,
                nasde_model=self.nasde_model)

            # determine the time since last storm
            stormDays, stormPrecip = storms.time_since_storm(
                self.precip,
                perc_snow,
                time_step=self.time_step/60/24,
                mass=self.ppt_threshold,
                time=self.time_to_end_storm,
                stormDays=self.storm_days,
                stormPrecip=self.storm_total)

            # save the model state
            self.percent_snow = perc_snow
            self.snow_density = snow_den
            self.storm_days = stormDays
            self.storm_total = stormPrecip

        else:

            self.storm_days += self.time_step/60/24

            # make everything else zeros
            self.precip = np.zeros(self.storm_days.shape)
            self.percent_snow = np.zeros(self.storm_days.shape)
            self.snow_density = np.zeros(self.storm_days.shape)

        # day of last storm, this will be used in albedo
        self.last_storm_day = utils.water_day(data.name)[0] - \
            self.storm_days - 0.001

    def distribute_thread(self, smrf_queue, data_queue):
        """
        Distribute the data using threading and queue. All data is provided and
        ``distribute_thread`` will go through each time step and call
        :mod:`smrf.distribute.precip.ppt.distribute` then puts the distributed
        data into the queue for:

        * :py:attr:`percent_snow`

        * :py:attr:`snow_density`
        * :py:attr:`storm_days`

        Args:
            queue: queue dictionary for all variables
            data: pandas dataframe for all data, indexed by date time
        """
        self._logger.info("Distributing {}".format(self.variable))

        for date_time in self.date_time:

            ppt_data = data_queue['precip'].get(date_time)
            ta_data = data_queue['air_temp'].get(date_time)
            ws_data = data_queue['wind_speed'].get(date_time)

            dpt = smrf_queue['dew_point'].get(date_time)
            precip_temp = smrf_queue['precip_temp'].get(date_time)
            ta = smrf_queue['air_temp'].get(date_time)

            # variables for wind redistribution
            if self.config['precip_rescaling_model'] == 'winstral':
                wind_direction = smrf_queue['wind_direction'].get(date_time)
                flatwind = smrf_queue['flatwind'].get(date_time)
                dir_round_cell = smrf_queue['dir_round_cell'].get(date_time)
                cell_maxus = smrf_queue['cellmaxus'].get(date_time)

            else:
                wind_direction = None
                flatwind = None
                dir_round_cell = None
                cell_maxus = None

            self.distribute(
                ppt_data,
                dpt,
                precip_temp,
                ta,
                date_time,
                ws_data,
                ta_data,
                wind_direction,
                dir_round_cell,
                flatwind,
                cell_maxus)

            smrf_queue[self.variable].put([date_time, self.precip])
            smrf_queue['percent_snow'].put([date_time, self.percent_snow])
            smrf_queue['snow_density'].put([date_time, self.snow_density])
            smrf_queue['storm_days'].put([date_time, self.storm_days])
            smrf_queue['storm_total'].put([date_time, self.storm_total])

            if self.nasde_model == "marks2017":
                smrf_queue['storm_id'].put([date_time, self.storm_id])

    def post_processor(self, main_obj, threaded=False):

        pass



import logging
import os
import shutil
import unittest
from copy import deepcopy
from pathlib import Path

import netCDF4 as nc
import numpy as np
from inicheck.tools import cast_all_variables, get_user_config

import smrf
from smrf.framework.model_framework import run_smrf


class SMRFTestCase(unittest.TestCase):
    """
    The base test case for SMRF that will load in the configuration file
    and store as the base config. Also will remove the output
    directory upon tear down.

    Runs the short simulation over reynolds mountain east
    """
    DIST_VARIABLES = frozenset([
        'air_temp',
        'cloud_factor',
        'precip',
        'thermal',
        'vapor_pressure',
        'wind',
    ])

    THREAD_CONFIG = {
        'threading': True,
        'max_queue': 1,
        'time_out': 5
    }

    BASE_INI_FILE_NAME = 'config.ini'

    test_dir = Path(smrf.__file__).parent.joinpath('tests')
    basin_dir = test_dir.joinpath('basins', 'RME')
    config_file = os.path.join(basin_dir, BASE_INI_FILE_NAME)

    @property
    def dist_variables(self):
        if self._dist_variables is None:
            self._dist_variables = list(self.DIST_VARIABLES)
        return self._dist_variables

    @property
    def base_config(self):
        return self.base_config_copy()

    @classmethod
    def base_config_copy(cls):
        return deepcopy(cls._base_config)

    @classmethod
    def load_base_config(cls):
        cls._base_config = get_user_config(cls.config_file, modules='smrf')

    @classmethod
    def setUpClass(cls):
        cls.load_base_config()
        cls.create_output_dir()

    @classmethod
    def tearDownClass(cls):
        cls.remove_output_dir()
        delattr(cls, 'output_dir')

    @classmethod
    def create_output_dir(cls):
        folder = os.path.join(cls._base_config.cfg['output']['out_location'])

        # Remove any potential files to ensure fresh run
        if os.path.isdir(folder):
            shutil.rmtree(folder)

        os.makedirs(folder)
        cls.output_dir = Path(folder)

    @classmethod
    def remove_output_dir(cls):
        if hasattr(cls, 'output_dir') and \
                os.path.exists(cls.output_dir):
            shutil.rmtree(cls.output_dir)

    @classmethod
    def thread_config(cls):
        config = cls.base_config_copy()
        config.raw_cfg['system'].update(cls.THREAD_CONFIG)

        config.apply_recipes()
        config = cast_all_variables(config, config.mcfg)

        return config

    @staticmethod
    def can_i_run_smrf(config):
        """
        Test whether a config is possible to run
        """
        try:
            run_smrf(config)
            return True

        except Exception as e:
            print(e)
            return False

    @staticmethod
    def assert_gold_equal(gold, not_gold, error_msg):
        """Compare two arrays

        Arguments:
            gold {array} -- gold array
            not_gold {array} -- not gold array
            error_msg {str} -- error message to display
        """

        if os.getenv('NOT_ON_GOLD_HOST') is None:
            try:
                np.testing.assert_array_equal(
                    not_gold,
                    gold,
                    err_msg=error_msg
                )
            except AssertionError:
                for rtol in [1e-5, 1e-4]:
                    status = np.allclose(
                        not_gold,
                        gold,
                        atol=0,
                        rtol=rtol
                    )
                    if status:
                        break
                if not status:
                    raise AssertionError(error_msg)
                print('Arrays are not exact match but close with rtol={}'.format(rtol))  # noqa
        else:
            np.testing.assert_almost_equal(
                not_gold,
                gold,
                decimal=3,
                err_msg=error_msg
            )

    def setUp(self):
        self._dist_variables = None

        # clear the logger
        for handler in logging.root.handlers:
            logging.root.removeHandler(handler)

    def tearDown(self):
        logging.shutdown()

    def compare_hrrr_gold(self):
        """
        Compare the model results with the gold standard
        """
        [
            self.compare_netcdf_files(file_name.name)
            for file_name in self.gold_dir.glob('*.nc')
        ]

    def compare_netcdf_files(self, output_file):
        """
        Compare two netcdf files to ensure that they are identical. The
        tests will compare the attributes of each variable and ensure that
        the values are exact
        """

        gold = nc.Dataset(self.gold_dir.joinpath(output_file))
        test = nc.Dataset(self.output_dir.joinpath(output_file))

        np.testing.assert_equal(
            gold.variables['time'][:],
            test.variables['time'][:],
            err_msg="Time steps did not match: \nGOLD {0} \n TEST {1}".format(
                gold.variables['time'], test.variables['time']
            )
        )

        # go through all variables and compare everything including
        # the attributes and data
        for var_name, v in gold.variables.items():

            # compare the dimensions
            for att in gold.variables[var_name].ncattrs():
                self.assertEqual(
                    getattr(gold.variables[var_name], att),
                    getattr(test.variables[var_name], att))

            # only compare those that are floats
            if gold.variables[var_name].datatype != np.dtype('S1'):
                error_msg = "Variable: {0} did not match gold standard". \
                    format(var_name)
                self.assert_gold_equal(
                    gold.variables[var_name][:],
                    test.variables[var_name][:],
                    error_msg
                )

        gold.close()
        test.close()



import numpy as np
from topocalc.horizon import horizon

from smrf.envphys.constants import (GRAVITY, IR_MAX, IR_MIN, MOL_AIR,
                                    SEA_LEVEL, STD_AIRTMP, STD_LAPSE,
                                    VISIBLE_MAX, VISIBLE_MIN)
from smrf.envphys.solar.irradiance import direct_solar_irradiance
from smrf.envphys.solar.twostream import twostream
from smrf.envphys.thermal.topotherm import hysat


def check_wavelengths(wavelength_range):

    if wavelength_range[0] >= VISIBLE_MIN and \
            wavelength_range[1] <= VISIBLE_MAX:
        wavelength_flag = 'vis'
    elif wavelength_range[0] >= IR_MIN and wavelength_range[1] <= IR_MAX:
        wavelength_flag = 'ir'
    else:
        raise ValueError(
            'stoporad wavelength range not within visible or IR wavelengths')

    return wavelength_flag


def stoporad(date_time, topo, cosz, azimuth, illum_ang, albedo_surface,
             wavelength_range, tau_elevation=100, tau=0.2, omega=0.85,
             scattering_factor=0.3):
    """[summary]

    Args:
        date_time ([type]): [description]
        topo ([type]): [description]
        cosz ([type]): [description]
        azimuth ([type]): [description]
        illum_ang ([type]): [description]
        albedo_surface ([type]): [description]
        wavelength_range ([type]): [description]
        tau_elevation (int, optional): [description]. Defaults to 100.
        tau (float, optional): [description]. Defaults to 0.2.
        omega (float, optional): [description]. Defaults to 0.85.
        scattering_factor (float, optional): [description]. Defaults to 0.3.

    Returns:
        [type]: [description]
    """

    wavelength_flag = check_wavelengths(wavelength_range)  # noqa

    # check cosz if sun is down
    if cosz < 0:
        return np.zeros_like(topo.dem), np.zeros_like(topo.dem)

    else:
        solar_irradiance = direct_solar_irradiance(
            date_time, w=wavelength_range)

        # Run horizon to get sun-below-horizon mask
        horizon_angles = horizon(azimuth, topo.dem, topo.dx)
        thresh = np.tan(np.pi / 2 - np.arccos(cosz))
        no_sun_mask = np.tan(np.abs(horizon_angles)) > thresh

        # Run shade to get cosine local illumination angle
        # mask by horizon mask using cosz=0 where the sun is not visible
        illum_ang = np.copy(illum_ang)
        illum_ang[no_sun_mask] = 0

        R0 = np.mean(albedo_surface)

        # Run elevrad to get beam & diffuse then toporad
        evrad = Elevrad(
            topo.dem,
            solar_irradiance,
            cosz,
            tau_elevation=tau_elevation,
            tau=tau,
            omega=omega,
            scattering_factor=scattering_factor,
            surface_albedo=R0)

        trad_beam, trad_diff = toporad(
            evrad.beam,
            evrad.diffuse,
            illum_ang,
            topo.sky_view_factor,
            topo.terrain_config_factor,
            cosz,
            surface_albedo=albedo_surface)

    return trad_beam, trad_diff


def toporad(beam, diffuse, illum_angle, sky_view_factor, terrain_config_factor,
            cosz, surface_albedo=0.0):
    """Topographically-corrected solar radiation. Calculates the topographic
    distribution of solar radiation at a single time, using input beam and
    diffuse radiation calculates supplied by elevrad.

    Args:
        beam (np.array): beam radiation
        diffuse (np.array): diffuse radiation
        illum_angle (np.array): local illumination angles
        sky_view_factor (np.array): sky view factor
        terrain_config_factor (np.array): terrain configuraiton factor
        cosz (float): cosine of the zenith
        surface_albedo (float/np.array, optional): surface albedo.
            Defaults to 0.0.

    Returns:
        tuple: beam and diffuse radiation corrected for terrain
    """

    # adjust diffuse radiation accounting for sky view factor
    drad = diffuse * sky_view_factor

    # add reflection from adjacent terrain
    drad = drad + (diffuse * (1 - sky_view_factor) +
                   beam * cosz) * terrain_config_factor * surface_albedo

    # global radiation is diffuse + incoming_beam * cosine of local
    # illumination * angle
    rad = drad + beam * illum_angle

    return rad, drad


class Elevrad():
    """Beam and diffuse radiation from elevation.
    elevrad is essentially the spatial or grid version of the twostream
    command.

    Args:
        elevation (np.array): DEM elevations in meters
        solar_irradiance (float): from direct_solar_irradiance
        cosz (float): cosine of zenith angle
        tau_elevation (float, optional): Elevation [m] of optical depth
                                        measurement. Defaults to 100.
        tau (float, optional): optical depth at tau_elevation. Defaults to 0.2.
        omega (float, optional): Single scattering albedo. Defaults to 0.85.
        scattering_factor (float, optional): Scattering asymmetry parameter.
                                            Defaults to 0.3.
        surface_albedo (float, optional): Mean surface albedo. Defaults to 0.5.
    """

    def __init__(self, elevation, solar_irradiance, cosz, **kwargs):
        """Initialize then run elevrad

        Args:
            elevation (np.array): DEM elevation in meters
            solar_irradiance (float): from direct_solar_irradiance
            cosz (float): cosine of zenith angle
            kwargs: tau_elevation, tau, omega, scattering_factor,
                    surface_albedo

        Returns:
            radiation: dict with beam and diffuse radiation
        """

        # defaults
        self.tau_elevation = 100.0
        self.tau = 0.2,
        self.omega = 0.85
        self.scattering_factor = 0.3
        self.surface_albedo = 0.5

        # set user specified values
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)

        self.elevation = elevation
        self.solar_irradiance = solar_irradiance
        self.cosz = cosz

        self.calculate()

    def calculate(self):
        """Perform the calculations
        """

        # reference pressure (at reference elevation, in km)
        reference_pressure = hysat(SEA_LEVEL, STD_AIRTMP, STD_LAPSE,
                                   self.tau_elevation / 1000, GRAVITY, MOL_AIR)

        # Convert each elevation in look-up table to pressure, then to optical
        # depth over the modeling domain
        pressure = hysat(SEA_LEVEL, STD_AIRTMP, STD_LAPSE,
                         self.elevation / 1000, GRAVITY, MOL_AIR)
        tau_domain = self.tau * pressure / reference_pressure

        # twostream over the optical depth of the domain
        self.twostream = twostream(
            self.cosz,
            self.solar_irradiance,
            tau=tau_domain,
            omega=self.omega,
            g=self.scattering_factor,
            R0=self.surface_albedo)

        # calculate beam and diffuse
        self.beam = self.solar_irradiance * \
            self.twostream['direct_transmittance']
        self.diffuse = self.solar_irradiance * self.cosz * \
            (self.twostream['transmittance'] -
             self.twostream['direct_transmittance'])



import numpy as np

from smrf.distribute import image_data
from smrf.envphys.constants import STEF_BOLTZ
from smrf.envphys.core import envphys_c
from smrf.envphys.thermal import clear_sky, cloud, vegetation
from smrf.utils import utils


class th(image_data.image_data):
    """
    The :mod:`~smrf.distribute.thermal.th` class allows for variable specific
    distributions that go beyond the base class.

    Thermal radiation, or long-wave radiation, is calculated based on the clear
    sky radiation emitted by the atmosphere. Multiple methods for calculating
    thermal radition exist and SMRF has 4 options for estimating clear sky
    thermal radiation. Selecting one of the options below will change the
    equations used. The methods were chosen based on the study by Flerchinger
    et al (2009) :cite:`Flerchinger&al:2009` who performed a model comparison
    using 21 AmeriFlux sites from North America and China.

    Marks1979
        The methods follow those developed by Marks and Dozier (1979)
        :cite:`Marks&Dozier:1979` that calculates the effective clear sky
        atmospheric emissivity using the distributed air temperature,
        distributed dew point temperature, and the elevation. The clear sky
        radiation is further adjusted for topographic affects based on the
        percent of the sky visible at any given point.

    Dilley1998
        .. math::
            L_{clear} = 59.38 + 113.7 * \\left( \\frac{T_a}{273.16} \\right)^6
            + 96.96 \\sqrt{w/25}

        References: Dilley and O'Brian (1998) :cite:`Dilley&OBrian:1998`

    Prata1996
        .. math::
            \\epsilon_{clear} = 1 - (1 + w) * exp(-1.2 + 3w)^{1/2}

        References: Prata (1996) :cite:`Prata:1996`

    Angstrom1918
        .. math::
            \\epsilon_{clear} = 0.83 - 0.18 * 10^{-0.067 e_a}

        References: Angstrom (1918) :cite:`Angstrom:1918` as cited by Niemela
        et al (2001) :cite:`Niemela&al:2001`

    .. figure:: ../_static/thermal_comparison.png
       :alt: Comparing the 4 thermal methods.

       The 4 different methods for estimating clear sky thermal radiation for a
       single time step. As compared to the Mark1979 method, the other methods
       provide a wide range in the estimated value of thermal radiation.


    The topographic correct clear sky thermal radiation is further adjusted for
    cloud affects. Cloud correction is based on fraction of cloud cover, a
    cloud factor close to 1 meaning no clouds are present, there is little
    radiation added.  When clouds are present, or a cloud factor close to 0,
    then additional long wave radiation is added to account for the cloud
    cover. Selecting one of the options below will change the equations used.
    The methods were chosen based on the study by Flerchinger et al (2009)
    :cite:`Flerchinger&al:2009`, where :math:`c=1-cloud\\_factor`.

    Garen2005
        Cloud correction is based on the relationship in Garen and Marks (2005)
        :cite:`Garen&Marks:2005` between the cloud factor and measured long
        wave radiation using measurement stations in the Boise River Basin.

        .. math::
            L_{cloud} = L_{clear} * (1.485 - 0.488 * cloud\\_factor)

    Unsworth1975
        .. math::

            L_d &= L_{clear} + \\tau_8 c f_8 \\sigma T^{4}_{c}

            \\tau_8 &= 1 - \\epsilon_{8z} (1.4 - 0.4 \\epsilon_{8z})

            \\epsilon_{8z} &= 0.24 + 2.98 \\times 10^{-6} e^2_o exp(3000/T_o)

            f_8 &= -0.6732 + 0.6240 \\times 10^{-2} T_c - 0.9140
            \\times 10^{-5} T^2_c

        References: Unsworth and Monteith (1975) :cite:`Unsworth&Monteith:1975`

    Kimball1982
        .. math::
            L_d &= L_{clear} + \\tau_8 c \\sigma T^4_c


        where the original Kimball et al. (1982) :cite:`Kimball&al:1982` was
        for multiple cloud layers, which was simplified to one layer.
        :math:`T_c` is the cloud temperature and is assumed to be 11 K cooler
        than :math:`T_a`.

        References: Kimball et al. (1982) :cite:`Kimball&al:1982`

    Crawford1999
        .. math::
            \\epsilon_a = (1 - cloud\\_factor) + cloud\\_factor *
            \\epsilon_{clear}

        References: Crawford and Duchon (1999) :cite:`Crawford&Duchon:1999`
        where :math:`cloud\\_factor` is the ratio of measured solar radiation
        to the clear sky irradiance.

    The results from Flerchinger et al (2009) :cite:`Flerchinger&al:2009`
    showed that the Kimball1982 cloud correction with Dilley1998 clear sky
    algorthim had the lowest RMSD. The Crawford1999 worked best when combined
    with Angstrom1918, Dilley1998, or Prata1996.

    .. figure:: ../_static/thermal_cloud_comparision.png
       :alt: Comparing the 4 thermal cloud correction methods.

       The 4 different methods for correcting clear sky thermal radiation for
       cloud affects at a single time step. As compared to the Garen2005
       method, the other methods are typically higher where clouds are present
       (i.e. the lower left) where the cloud factor is around 0.4.

    The thermal radiation is further adjusted for canopy cover after the work
    of Link and Marks (1999) :cite:`Link&Marks:1999`. The correction is based
    on the vegetation's transmissivity, with the canopy temperature assumed to
    be the air temperature for vegetation greater than 2 meters.  The thermal
    radiation is adjusted by

    .. math::
        L_{canopy} = \\tau_d * L_{cloud} + (1 - \\tau_d) \\epsilon
        \\sigma T_a^4

    where :math:`\\tau_d` is the optical transmissivity, :math:`L_{cloud}` is
    the cloud corrected thermal radiation, :math:`\\epsilon` is the emissivity
    of the canopy (0.96), :math:`\\sigma` is the Stephan-Boltzmann constant,
    and :math:`T_a` is the distributed air temperature.

    Args:
        thermalConfig: The [thermal] section of the configuration file

    Attributes:
        config: configuration from [thermal] section
        thermal: numpy array of the precipitation
        min: minimum value of thermal is -600 W/m^2
        max: maximum value of thermal is 600 W/m^2
        stations: stations to be used in alphabetical order
        dem: numpy array for the DEM, from
            :py:attr:`smrf.data.loadTopo.Topo.dem`
        veg_type: numpy array for the veg type, from
            :py:attr:`smrf.data.loadTopo.Topo.veg_type`
        veg_height: numpy array for the veg height, from
            :py:attr:`smrf.data.loadTopo.Topo.veg_height`
        veg_k: numpy array for the veg K, from
            :py:attr:`smrf.data.loadTopo.Topo.veg_k`
        veg_tau: numpy array for the veg transmissivity, from
            :py:attr:`smrf.data.loadTopo.Topo.veg_tau`
        sky_view_factor: numpy array for the sky view factor, from
            :py:attr:`smrf.data.loadTopo.Topo.sky_view_factor`
    """

    variable = 'thermal'

    # these are variables that can be output
    OUTPUT_VARIABLES = {
        'thermal': {
            'units': 'watt/m2',
            'standard_name': 'thermal_radiation',
            'long_name': 'Thermal (longwave) radiation'
        },
        'thermal_clear': {
            'units': 'watt/m2',
            'standard_name': 'thermal_radiation non-correct',
            'long_name': 'Thermal (longwave) radiation non-corrected'
        },
        'thermal_cloud': {
            'units': 'watt/m2',
            'standard_name': 'thermal_radiation cloud corrected',
            'long_name': 'Thermal (longwave) radiation cloud corrected'
        },
        'thermal_veg': {
            'units': 'watt/m2',
            'standard_name': 'thermal_radiation veg corrected',
            'long_name': 'Thermal (longwave) radiation veg corrected'
        }
    }
    # these are variables that are operate at the end only and do not need to
    # be written during main distribute loop
    post_process_variables = {}

    BASE_THREAD_VARIABLES = frozenset([
        'thermal',
        'thermal_clear'
    ])

    def __init__(self, thermalConfig):

        # extend the base class
        image_data.image_data.__init__(self, self.variable)
        self.getConfig(thermalConfig)

        self.min = thermalConfig['min']
        self.max = thermalConfig['max']

        self.correct_cloud = self.config['correct_cloud']
        self.correct_veg = self.config['correct_veg']
        self.correct_terrain = self.config['correct_terrain']

        if self.correct_cloud:
            self.cloud_method = self.config['cloud_method']

        self.clear_sky_method = self.config['clear_sky_method']

        self._logger.debug('Created distribute.thermal')

    def initialize(self, topo, data, date_time=None):
        """
        Initialize the distribution, calls
        :mod:`smrf.distribute.image_data.image_data._initialize` for gridded
        distirbution. Sets the following from :mod:`smrf.data.loadTopo.Topo`

        * :py:attr:`veg_height`
        * :py:attr:`veg_tau`
        * :py:attr:`veg_k`
        * :py:attr:`sky_view_factor`
        * :py:attr:`dem`

        Args:
            topo: :mod:`smrf.data.loadTopo.Topo` instance contain topographic
                data and infomation
            data: data Pandas dataframe containing the station data,
                from :mod:`smrf.data.loadData` or :mod:`smrf.data.loadGrid`

        """

        self._logger.debug('Initializing distribute.thermal')
        self.date_time = date_time
        self._initialize(topo, data.metadata)

        self.veg_height = topo.veg_height
        self.veg_tau = topo.veg_tau
        self.veg_k = topo.veg_k
        self.sky_view_factor = topo.sky_view_factor
        if not self.correct_terrain:
            self.sky_view_factor = None
        self.dem = topo.dem

        if self.correct_cloud:
            self.add_thread_variables('thermal_cloud')

        if self.correct_veg:
            self.add_thread_variables('thermal_veg')

    def distribute(self, date_time, air_temp, vapor_pressure=None,
                   dew_point=None, cloud_factor=None):
        """
        Distribute for a single time step.

        The following steps are taken when distributing thermal:

        1. Calculate the clear sky thermal radiation from
            :mod:`smrf.envphys.core.envphys_c.ctopotherm`
        2. Correct the clear sky thermal for the distributed cloud factor
        3. Correct for canopy affects

        Args:
            date_time: datetime object for the current step
            air_temp: distributed air temperature for the time step
            vapor_pressure: distributed vapor pressure for the time step
            dew_point: distributed dew point for the time step
            cloud_factor: distributed cloud factor for the time step
                measured/modeled
        """

        self._logger.debug('%s Distributing thermal' % date_time)

        # calculate clear sky thermal
        if self.clear_sky_method == 'marks1979':
            cth = np.zeros_like(air_temp, dtype=np.float64)
            envphys_c.ctopotherm(
                air_temp, dew_point,
                self.dem,
                self.sky_view_factor,
                cth,
                self.config['marks1979_nthreads'])

        elif self.clear_sky_method == 'dilley1998':
            cth = clear_sky.Dilly1998(air_temp, vapor_pressure/1000)

        elif self.clear_sky_method == 'prata1996':
            cth = clear_sky.Prata1996(air_temp, vapor_pressure/1000)

        elif self.clear_sky_method == 'angstrom1918':
            cth = clear_sky.Angstrom1918(air_temp, vapor_pressure/1000)

        # terrain factor correction
        if (self.sky_view_factor is not None) and \
                (self.clear_sky_method != 'marks1979'):
            # apply (emiss * skvfac) + (1.0 - skvfac) to the longwave
            cth = cth * self.sky_view_factor + (1.0 - self.sky_view_factor) * \
                STEF_BOLTZ * air_temp**4

        # make output variable
        self.thermal_clear = cth.copy()

        # correct for the cloud factor
        # ratio of measured/modeled solar indicates the thermal correction
        if self.correct_cloud:
            if self.cloud_method == 'garen2005':
                cth = cloud.Garen2005(cth,
                                      cloud_factor)

            elif self.cloud_method == 'unsworth1975':
                cth = cloud.Unsworth1975(cth,
                                         air_temp,
                                         cloud_factor)

            elif self.cloud_method == 'kimball1982':
                cth = cloud.Kimball1982(cth,
                                        air_temp,
                                        vapor_pressure/1000,
                                        cloud_factor)

            elif self.cloud_method == 'crawford1999':
                cth = cloud.Crawford1999(cth,
                                         air_temp,
                                         cloud_factor)

            # make output variable
            self.thermal_cloud = cth.copy()

        # correct for vegetation
        if self.correct_veg:
            cth = vegetation.thermal_correct_canopy(cth,
                                                    air_temp,
                                                    self.veg_tau,
                                                    self.veg_height)

            # make output variable
            self.thermal_veg = cth.copy()

        self.thermal = utils.set_min_max(cth, self.min, self.max)

    def distribute_thread(self, smrf_queue, data_queue=None):
        """
        Distribute the data using threading. All data is provided and
        ``distribute_thread`` will go through each time step and call
        :mod:`smrf.distribute.thermal.th.distribute` then puts the distributed
        data into the smrf_queue for :py:attr:`thermal`.

        Args:
            smrf_queue: smrf_queue dictionary for all variables
            data: pandas dataframe for all data, indexed by date time

        """

        for date_time in self.date_time:

            air_temp = smrf_queue['air_temp'].get(date_time)
            dew_point = smrf_queue['dew_point'].get(date_time)
            vapor_pressure = smrf_queue['vapor_pressure'].get(date_time)
            cloud_factor = smrf_queue['cloud_factor'].get(date_time)

            self.distribute(date_time, air_temp, vapor_pressure,
                            dew_point, cloud_factor)

            if self.correct_veg:
                smrf_queue['thermal_veg'].put(
                    [date_time, self.thermal_veg])
            if self.correct_cloud:
                smrf_queue['thermal_cloud'].put(
                    [date_time, self.thermal_cloud])

            smrf_queue['thermal_clear'].put([date_time, self.thermal_clear])

            smrf_queue['thermal'].put([date_time, self.thermal])



VISIBLE_MIN = .28
VISIBLE_MAX = .7
IR_MIN = .7
IR_MAX = 2.8
# visible solar irradiance wavelengths
VISIBLE_WAVELENGTHS = [VISIBLE_MIN, VISIBLE_MAX]
# infrared solar irradiance wavelengths
IR_WAVELENGTHS = [IR_MIN, IR_MAX]
SOLAR_CONSTANT = 1368.0  # solar constant in W/m**2

MAXV = 1.0              # vis albedo when gsize = 0
MAXIR = 0.85447         # IR albedo when gsize = 0
IRFAC = -0.02123        # IR decay factor
VFAC = 500.0            # visible decay factor
VZRG = 1.375e-3         # vis zenith increase range factor
IRZRG = 2.0e-3          # ir zenith increase range factor
IRZ0 = 0.1              # ir zenith increase range, gsize=0
STEF_BOLTZ = 5.6697e-8  # stephman boltzman constant
EMISS_TERRAIN = 0.98    # emissivity of the terrain
EMISS_VEG = 0.96        # emissivity of the vegitation
FREEZE = 273.16         # freezing temp K
BOIL = 373.15           # boiling temperature K
STD_LAPSE_M = -0.0065   # lapse rate (K/m)
STD_LAPSE = -6.5        # lapse rate (K/km)
SEA_LEVEL = 1.013246e5  # sea level pressure
RGAS = 8.31432e3        # gas constant (J / kmole / deg)
GRAVITY = 9.80665       # gravity (m/s^2)
MOL_AIR = 28.9644       # molecular weight of air (kg / kmole)
STD_AIRTMP = 2.88e2



import logging

import numpy as np
import pandas as pd
from weather_forecast_retrieval.hrrr import HRRR

from smrf.envphys.solar.cloud import get_hrrr_cloud
from smrf.envphys.vapor_pressure import rh2vp


class InputGribHRRR():

    DATA_TYPE = 'hrrr_grib'

    VARIABLES = [
        'air_temp',
        'vapor_pressure',
        'precip',
        'wind_speed',
        'wind_direction',
        'cloud_factor'
    ]

    def __init__(self, start_date, end_date, bbox=None,
                 topo=None, config=None):

        self.start_date = start_date
        self.end_date = end_date
        self.topo = topo
        self.bbox = bbox
        self.config = config
        self.time_zone = start_date.tzinfo

        if topo is None:
            raise Exception('Must supply topo to InputWRF')

        if bbox is None:
            raise Exception('Must supply bbox to InputWRF')

        self._logger = logging.getLogger(__name__)

        if self.config['hrrr_load_method'] == 'timestep':
            self._timedelta_steps = pd.to_timedelta(20, 'minutes')
            self.timestep_dates()
            self.cf_memory = None

        self.hrrr = HRRR(external_logger=self._logger)

    def timestep_dates(self):
        self.end_date = self.start_date + self._timedelta_steps

    def load(self):
        """
        Load the data from the High Resolution Rapid Refresh (HRRR) model
        The variables returned from the HRRR class in dataframes are

            - metadata
            - air_temp
            - relative_humidity
            - precip_int
            - cloud_factor
            - wind_u
            - wind_v

        The function will take the keys and load them into the appropriate
        objects within the `grid` class. The vapor pressure will be calculated
        from the `air_temp` and `relative_humidity`. The `wind_speed` and
        `wind_direction` will be calculated from `wind_u` and `wind_v`
        """

        self._logger.info('Reading data from from HRRR directory: {}'.format(
            self.config['hrrr_directory']
        ))

        metadata, data = self.hrrr.get_saved_data(
            self.start_date,
            self.end_date,
            self.bbox,
            output_dir=self.config['hrrr_directory'],
            force_zone_number=self.topo.zone_number
        )

        self.parse_data(metadata, data)

    def load_timestep(self, date_time):
        """Load a single time step for HRRR

        Args:
            date_time (datetime): date time to load
        """

        self.start_date = date_time
        self.timestep_dates()
        self.load()
        self.check_cloud_factor()

    def load_timestep_thread(self, date_times, data_queue):
        """Load HRRR within a thread and add the data to the data
        queue.

        Args:
            date_times (list): list of the simulation date_times
            data_queue (dict): dict of the data queues
        """

        for date_time in date_times[1:]:
            self.load_timestep(date_time)

            for variable in self.VARIABLES:
                data_queue[variable].put(
                    [date_time, getattr(self, variable).iloc[0]])

        self._logger.debug('Finished loading data')

    def parse_data(self, metadata, data):
        """Parse the data from HRRR into dataframes for SMRF

        Args:
            metadata (DataFrame): metadata DataFrame
            data (dict): dictionary of DataFrames from HRRR
        """

        # the data may be returned as type=object, convert to numeric
        # correct for the timezone
        for key in data.keys():
            data[key] = data[key].apply(pd.to_numeric)
            data[key] = data[key].tz_localize(tz=self.time_zone)

        self.metadata = metadata

        idx = data['air_temp'].index
        cols = data['air_temp'].columns

        self._logger.debug('Loading air_temp')
        self.air_temp = data['air_temp']

        # calculate vapor pressure
        self._logger.debug('Calculating vapor_pressure')
        vp = rh2vp(
            data['air_temp'].values,
            data['relative_humidity'].values)
        self.vapor_pressure = pd.DataFrame(vp, index=idx, columns=cols)

        # calculate the wind speed and wind direction
        self._logger.debug('Loading wind_speed and wind_direction')

        s = np.sqrt(data['wind_u']**2 + data['wind_v']**2)

        d = np.degrees(np.arctan2(data['wind_v'], data['wind_u']))
        ind = d < 0
        d[ind] = d[ind] + 360
        self.wind_speed = pd.DataFrame(s, index=idx, columns=cols)
        self.wind_direction = pd.DataFrame(d, index=idx, columns=cols)

        # precip
        self._logger.debug('Loading precip')
        self.precip = pd.DataFrame(data['precip_int'], index=idx, columns=cols)

        # cloud factor
        self._logger.debug('Loading solar')
        solar = pd.DataFrame(data['short_wave'], index=idx, columns=cols)
        self._logger.debug('Calculating cloud factor')
        self.cloud_factor = get_hrrr_cloud(
            solar, self.metadata,
            self.topo.basin_lat, self.topo.basin_long)

    def check_cloud_factor(self):
        """Check the cloud factor when in the timestep mode.
        This will fill NaN values as they happen by linearly
        interpolating from the last hour (i.e. copy it). This
        is similar to how `get_hrrr_cloud` with the difference
        being that it won't interpolate over the entire night.
        """

        if self.cf_memory is None:
            self.cf_memory = self.cloud_factor

            # if initial cloud factor is at night, default to clear sky
            if self.cf_memory.isnull().values.any():
                self.cf_memory[:] = 1

        else:
            self.cf_memory = pd.concat(
                [self.cf_memory.tail(1), self.cloud_factor])

        self.cf_memory = self.cf_memory.interpolate(method='linear').ffill()

        if self.cf_memory.isnull().values.any():
            self._logger.error('There are NaN values in the cloud factor')

        self.cloud_factor = self.cf_memory.tail(1)



# cython: embedsignature=True
# cython: language_level=3
'''
Compiling dk's kriging function

20160205 Scott Havens
'''


import cython
import numpy as np

cimport numpy as np
from cpython cimport Py_INCREF, PyObject
from libc.stdlib cimport free

# Numpy must be initialized. When using numpy from C or Cython you must
# _always_ do that, or you will have segfaults
np.import_array()


cdef extern from "dk_header.h":
    double *krige(int nsta, double *ad, double *dgrid, double *elevations);
    void krige_grid(int nsta, int ngrid, double *ad, double *dgrid, double *elevations, int nthreads, double *weights);
#     int lusolv(int n, double **a, double *x)



@cython.boundscheck(False)
@cython.wraparound(False)
# https://github.com/cython/cython/wiki/tutorials-NumpyPointerToC
def call_grid(object ad, object dgrid,
              np.ndarray[double, mode="c", ndim=1] elevations, 
              np.ndarray[double, mode="c", ndim=2] weights not None,
              int nthreads=1):
    '''
    Call the function krige_grid in krige.c which will iterate over the grid
    within the C code

    Args:
        ad - [nsta x nsta] matrix of distances between stations
        dgrid - [ngrid x nsta] matrix of distances between grid points and stations
        elevations - [nsta] array of station elevations
        weights (return) - [ngrid x nsta] matrix of kriging weights calculated
        nthreads - number of threads to use in parallel processing

    Out:
        weights changed in place

    20160222 Scott Havens
    '''

    cdef int nsta, ngrid
    ngrid, nsta = dgrid.shape[0], dgrid.shape[1]

    # convert the ad array to C
    cdef np.ndarray[double, mode="c", ndim=2] ad_arr
    ad_arr = np.ascontiguousarray(ad, dtype=np.float64)

    # convert the dgrid to C
    cdef np.ndarray[double, mode="c", ndim=2] grid
    grid = np.ascontiguousarray(dgrid, dtype=np.float64)

    # call the C function
    krige_grid(nsta, ngrid, &ad_arr[0,0], &grid[0,0], &elevations[0], nthreads, &weights[0,0])

    return None


# @cython.boundscheck(False)
# @cython.wraparound(False)
# def run(object ad, object dgrid,
#                np.ndarray[double, mode="c", ndim=1] elevations,
#                np.ndarray[double, mode="c", ndim=2] weights not None):
#     '''
#     Instead of calling the krige_grid, iterate over the grid here and
#     add parallel support
#
#     This isn't working or finished
#     '''
#
#     cdef double *w
#     cdef np.ndarray w_arr
#
#     cdef int nsta, ngrid
#     ngrid, nsta = dgrid.shape[0], dgrid.shape[1]
#
#     # convert the ad array to C
#     cdef np.ndarray[double, mode="c", ndim=2] ad_arr
#     ad_arr = np.ascontiguousarray(ad, dtype=np.float64)
#
#     # convert the dgrid to C
#     cdef np.ndarray[double, mode="c", ndim=2] grid
#     grid = np.ascontiguousarray(dgrid, dtype=np.float64)
#
#     for i in range(ngrid):
#
#         w = krige(nsta, &ad_arr[0,0], &grid[i,0], &elevations[0]);
#
#         # convert C array to Python object
#         array_wrapper = ArrayWrapper()
#         array_wrapper.set_data(nsta, <void*> w)
#         w_arr = np.array(array_wrapper, copy=False)
#         # Assign our object to the 'base' of the ndarray object
#         w_arr.base = <PyObject*> array_wrapper
#         # Increment the reference count, as the above assignement was done in
#         # C, and Python does not know that there is this additional reference
#         Py_INCREF(array_wrapper)
#
#
#         weights[i] = w_arr
#
#     return None


# @cython.boundscheck(False)
# @cython.wraparound(False)
# def call_func(int l, int nsta, object ad,
#                np.ndarray[double, mode="c", ndim=1] dgrid,
#                np.ndarray[double, mode="c", ndim=1] elevations):
#
#     cdef double *array
#     cdef np.ndarray ndarray
#
#     cdef np.ndarray[double, mode="c", ndim=2] arr
#     arr = np.ascontiguousarray(ad, dtype=np.float64)
#
#
#     print 'here python'
#
# #     cdef a = <double**> ad.data
#
#     # call the C function
#     array = krige(nsta, &arr[0,0], &dgrid[0], &elevations[0])
#
# #     array = krige(nsta, <double**> arr.data, &dgrid[0], &elevations[0])
#
#
#
#
# #     # convert C array to Python object
# #     array_wrapper = ArrayWrapper()
# #     array_wrapper.set_data(nsta, <void*> array)
# #     ndarray = np.array(array_wrapper, copy=False)
# #     # Assign our object to the 'base' of the ndarray object
# #     ndarray.base = <PyObject*> array_wrapper
# #     # Increment the reference count, as the above assignement was done in
# #     # C, and Python does not know that there is this additional reference
# #     Py_INCREF(array_wrapper)
#
#
#     return ndarray

# We need to build an array-wrapper class to deallocate our array when
# the Python object is deleted.
# https://gist.github.com/GaelVaroquaux/1249305

# cdef class ArrayWrapper:
#     cdef void* data_ptr
#     cdef int size
#
#     cdef set_data(self, int size, void* data_ptr):
#         """ Set the data of the array
#         This cannot be done in the constructor as it must recieve C-level
#         arguments.
#         Parameters:
#         -----------
#         size: int
#             Length of the array.
#         data_ptr: void*
#             Pointer to the data
#         """
#         self.data_ptr = data_ptr
#         self.size = size
#
#     def __array__(self):
#         """ Here we use the __array__ method, that is called when numpy
#             tries to get an array from the object."""
#         cdef np.npy_intp shape[1]
#         shape[0] = <np.npy_intp> self.size
#         # Create a 1D array, of length 'size'
#         ndarray = np.PyArray_SimpleNewFromData(1, shape,
#                                                np.NPY_INT, self.data_ptr)
#         return ndarray
#  
#     def __dealloc__(self):
#         """ Frees the array. This is called by Python when all the
#         references to the object are gone. """
#         free(<void*>self.data_ptr)


from copy import deepcopy

from inicheck.tools import cast_all_variables

from smrf.framework.model_framework import run_smrf
from smrf.tests.smrf_test_case import SMRFTestCase


class TestLoadNetcdf(SMRFTestCase):

    def test_grid_netcdf(self):

        config = deepcopy(self.base_config)
        del config.raw_cfg['csv']

        generic_grid = {
            'data_type': 'netcdf',
            'netcdf_file': './gridded/netcdf_test.nc',
            'air_temp': 'air_temp',
            'vapor_pressure': 'vapor_pressure',
            'precip': 'precip',
            'wind_speed': 'wind_speed',
            'wind_direction': 'wind_direction',
            'cloud_factor': 'cloud_factor'
        }
        config.raw_cfg['gridded'] = generic_grid
        config.raw_cfg['system']['time_out'] = '25'
        config.raw_cfg['system']['queue_max_values'] = '2'
        # Doesn't work with true
        config.raw_cfg['system']['threading'] = 'False'

        # set the distribution to grid, thermal defaults will be fine
        for v in self.dist_variables:
            config.raw_cfg[v]['distribution'] = 'grid'
            config.raw_cfg[v]['grid_mask'] = 'False'

        config.raw_cfg['thermal']['correct_cloud'] = 'False'
        config.raw_cfg['thermal']['correct_veg'] = 'True'

        config.raw_cfg['wind']['wind_model'] = 'interp'

        # fix the time to that of the WRF_test.nc
        config.raw_cfg['time']['start_date'] = '2015-03-03 00:00'
        config.raw_cfg['time']['end_date'] = '2015-03-03 04:00'

        config.apply_recipes()
        config = cast_all_variables(config, config.mcfg)

        # ensure that the recipes are used
        self.assertTrue(
            'station_adjust_for_undercatch' not in config.cfg['precip'].keys())
        self.assertFalse(config.cfg['thermal']['correct_cloud'])
        self.assertTrue(config.cfg['thermal']['correct_veg'])

        run_smrf(config)



import logging

import netCDF4 as nc
import numpy as np
import pandas as pd

from smrf.distribute import image_data
from smrf.utils import utils


class WinstralWindModel(image_data.image_data):
    """Estimating wind speed and direction is complex terrain can be difficult due
    to the interaction of the local topography with the wind. The methods
    described here follow the work developed by Winstral and Marks (2002) and
    Winstral et al. (2009) :cite:`Winstral&Marks:2002` :cite:`Winstral&al:2009`
    which parameterizes the terrain based on the upwind direction. The
    underlying method calulates the maximum upwind slope (maxus) within a
    search distance to determine if a cell is sheltered or exposed. See
    :mod:`smrf.utils.wind.model` for a more in depth description. A maxus file
    (library) is used to load the upwind direction and maxus values over the
    dem. The following steps are performed when estimating the wind speed:

    1. Adjust measured wind speeds at the stations and determine the wind
       direction componenets
    2. Distribute the flat wind speed
    3. Distribute the wind direction components
    4. Simulate the wind speeds based on the distribute flat wind, wind
       direction, and maxus values

    After the maxus is calculated for multiple wind directions over the entire
    DEM, the measured wind speed and direction can be distirbuted. The first
    step is to adjust the measured wind speeds to estimate the wind speed if
    the site were on a flat surface. The adjustment uses the maxus value at the
    station location and an enhancement factor for the site based on the
    sheltering of that site to wind. A special consideration is performed when
    the station is on a peak, as artificially high wind speeds can be
    calcualted.  Therefore, if the station is on a peak, the minimum maxus
    value is choosen for all wind directions. The wind direction is also broken
    up into the u,v componenets.

    Next the flat wind speed, u wind direction component, and v wind direction
    compoenent are distributed using the underlying distribution methods. With
    the distributed flat wind speed and wind direction, the simulated wind
    speeds can be estimated. The distributed wind direction is binned into the
    upwind directions in the maxus library. This determines which maxus value
    to use for each pixel in the DEM. Each cell's maxus value is further
    enhanced for vegetation, with larger, more dense vegetation increasing the
    maxus value (more sheltering) and bare ground not enhancing the maxus value
    (exposed). With the adjusted maxus values, wind speed is estimated using
    the relationships in Winstral and Marks (2002) and Winstral et al. (2009)
    :cite:`Winstral&Marks:2002` :cite:`Winstral&al:2009` based on the
    distributed flat wind speed and each cell's maxus value.
    """

    VARIABLE = 'wind'

    BASE_THREAD_VARIABLES = frozenset([
        'flatwind',
        'cellmaxus',
        'dir_round_cell'
    ])

    def __init__(self, smrf_config):
        """Initialize the WinstralWindModel

        Arguments:
            smrf_config {UserConfig} -- entire smrf config
            distribute_drifts {bool} -- distribute drifts if true

        Raises:
            IOError: if maxus file does not match topo size
        """

        image_data.image_data.__init__(self, self.VARIABLE)

        self._logger = logging.getLogger(__name__)

        self.smrf_config = smrf_config
        self.getConfig(smrf_config['wind'])
        # self.distribute_drifts = distribute_drifts

        self._logger.debug('Creating the WinstralWindModel')

        # open the maxus netCDF
        self._maxus_file = nc.Dataset(self.config['maxus_netcdf'], 'r')
        self.maxus = self._maxus_file.variables['maxus'][:]
        self.maxus_direction = self._maxus_file.variables['direction'][:]
        self._maxus_file.close()

        # Maxus must be the the same size as the topo.
        topo_nc = nc.Dataset(self.smrf_config['topo']['filename'], 'r')
        t_shape = topo_nc.variables['dem'].shape
        topo_nc.close()

        if t_shape != self.maxus[0].shape:
            raise IOError("\nMaxus file must be generated using the topo to"
                          " be valid. Maxus netcdf shape = {} and topo"
                          " netcdf shape = {}".format(t_shape,
                                                      self.maxus.shape))

        self._logger.debug('Read data from {}'
                           .format(self.config['maxus_netcdf']))

        # get the veg values
        matching = [s for s in self.config.keys() if "veg_" in s]
        v = {}
        for m in matching:
            ms = m.split('_')
            if type(self.config[m]) == list:
                v[ms[1]] = float(self.config[m][0])
            else:
                v[ms[1]] = float(self.config[m])

        self.veg = v

    def initialize(self, topo, data):
        """Initialize the model with data

        Arguments:
            topo {topo class} -- Topo class
            data {data object} -- SMRF data object
        """

        self._logger.debug('Initializing the WinstralWindModel')

        self._initialize(topo, data.metadata)

        self.veg_type = topo.veg_type

        # meshgrid points
        self.X = topo.X
        self.Y = topo.Y

        # get the enhancements for the stations
        if 'enhancement' not in self.metadata.columns:
            self.metadata['enhancement'] = \
                float(self.config['station_default'])

            for m in self.metadata.index:
                sta_e = m.lower()
                if sta_e in self.config:
                    if type(self.config[sta_e]) == list:
                        enhancement = self.config[sta_e][0]
                    else:
                        enhancement = self.config[sta_e]

                    self.metadata.loc[m, 'enhancement'] = \
                        float(enhancement)

        # if not self.distribute_drifts:
        # we have to pass these to precip, so make them none
        # if we won't use them, or they will be overwritten later
        # self.dir_round_cell = None
        # self.cellmaxus = None

    def distribute(self, data_speed, data_direction):
        """Distribute the wind for the model

        Follows the following steps for station measurements:

        1. Adjust measured wind speeds at the stations and determine the wind
            direction componenets
        2. Distribute the flat wind speed
        3. Distribute the wind direction components
        4. Simulate the wind speeds based on the distribute flat wind, wind
            direction, and maxus values

        Arguments:
            data_speed {DataFrame} -- wind speed data frame
            data_direction {DataFrame} -- wind direction data frame
        """

        # calculate the maxus at each site
        self.stationMaxus(data_speed, data_direction)

        # distribute the flatwind
        self._distribute(self.flatwind_point,
                         other_attribute='flatwind')

        # distribute u_direction and v_direction
        self._distribute(self.u_direction,
                         other_attribute='u_direction_distributed')
        self._distribute(self.v_direction,
                         other_attribute='v_direction_distributed')

        # Calculate simulated wind speed at each cell from flatwind
        self.simulateWind(data_speed)

    def simulateWind(self, data_speed):
        """
        Calculate the simulated wind speed at each cell from flatwind and the
        distributed directions. Each cell's maxus value is pulled from the
        maxus library based on the distributed wind direction. The cell's maxus
        is further adjusted based on the vegetation type and the factors
        provided in the [wind] section of the configuration file.

        Args:
            data_speed: Pandas dataframe for a single time step of wind speed
                to make the pixel locations same as the measured values
        """

        # combine u and v to azimuth
        az = np.arctan2(self.u_direction_distributed,
                        self.v_direction_distributed)*180/np.pi
        az[az < 0] = az[az < 0] + 360

        dir_round_cell = np.ceil((az - self.nstep/2) / self.nstep) * self.nstep
        dir_round_cell[dir_round_cell <
                       0] = dir_round_cell[dir_round_cell < 0] + 360
        dir_round_cell[dir_round_cell == -0] = 0
        dir_round_cell[dir_round_cell == 360] = 0

        cellmaxus = np.zeros(dir_round_cell.shape)
        cellwind = np.zeros(dir_round_cell.shape)

        dir_unique = np.unique(dir_round_cell)
        for d in dir_unique:
            # find all values for matching direction
            ind = dir_round_cell == d
            i = np.argwhere(self.maxus_direction == d)[0][0]
            cellmaxus[ind] = self.maxus[i][ind]

        # correct for veg
        dynamic_mask = np.ones(cellmaxus.shape)
        for k, v in self.veg.items():
            # Adjust veg types that were specified by the user
            if k != 'default':
                ind = self.veg_type == int(k)
                dynamic_mask[ind] = 0
                cellmaxus[ind] += v

        # Apply the veg default to those that weren't messed with
        if self.veg['default'] != 0:
            cellmaxus[dynamic_mask == 1] += self.veg['default']

        # correct unreasonable values
        cellmaxus[cellmaxus > 32] = 32
        cellmaxus[cellmaxus < -32] = -32

        # determine wind
        factor = float(self.config['reduction_factor'])
        ind = cellmaxus < -30.10
        cellwind[ind] = factor * self.flatwind[ind] * 4.211

        ind = (cellmaxus > -30.10) & (cellmaxus < -21.3)
        c = np.abs(cellmaxus[ind])
        cellwind[ind] = factor * self.flatwind[ind] * \
            (1.756507 - 0.1678945 * c + 0.01927844 * np.power(c, 2) -
             0.0003651592 * np.power(c, 3))

        ind = (cellmaxus > -21.3) & (cellmaxus < 0)
        c = np.abs(cellmaxus[ind])
        cellwind[ind] = factor * self.flatwind[ind] * \
            (1.0 + 0.1031717 * c - 0.008003561 * np.power(c, 2) +
             0.0003996581 * np.power(c, 3))

        ind = cellmaxus > 30.10
        cellwind[ind] = self.flatwind[ind] / 4.211

        ind = (cellmaxus < 30.10) & (cellmaxus > 21.3)
        c = cellmaxus[ind]
        cellwind[ind] = self.flatwind[ind] / \
            (1.756507 - 0.1678945 * c + 0.01927844 * np.power(c, 2) -
             0.0003651592 * np.power(c, 3))

        ind = (cellmaxus < 21.3) & (cellmaxus >= 0)
        c = cellmaxus[ind]
        cellwind[ind] = self.flatwind[ind] / \
            (1.0 + 0.1031717 * c - 0.008003561 * np.power(c, 2) +
             0.0003996581 * np.power(c, 3))

        # Convert from 3m to 5m wind speed
        cellwind *= 1.07985

        # preseve the measured values
        cellwind[self.metadata.yi, self.metadata.xi] = data_speed

        # check for NaN
        nans, x = utils.nan_helper(cellwind)

        if np.sum(nans) > 0:
            cellwind[nans] = np.interp(x(nans), x(~nans), cellwind[~nans])

        self.wind_speed = utils.set_min_max(cellwind, self.min, self.max)
        self.wind_direction = az
        self.cellmaxus = cellmaxus
        self.dir_round_cell = dir_round_cell

    def stationMaxus(self, data_speed, data_direction):
        """
        Determine the maxus value at the station given the wind direction.
        Can specify the enhancemet for each station or use the default, along
        with whether or not the station is on a peak which will ensure that
        the station cannot be sheltered. The station enhancement and peak
        stations are specified in the [wind] section of the configuration
        file. Calculates the following for each station:

        * :py:attr:`flatwind`
        * :py:attr:`u_direction`
        * :py:attr:`v_direction`

        Args:
            data_speed: wind_speed data frame for single time step
            data_direction: wind_direction data frame for single time step

        """

        # ----------------------------------------
        # Get data and site maxus value
        flatwind = data_speed.copy()

        # number of bins that the maxus library was calculated for
        self.nbins = len(self.maxus_direction)
        self.nstep = 360/self.nbins

        for m in self.metadata.index:
            # pixel locations
            xi = self.metadata.loc[m, 'xi']
            yi = self.metadata.loc[m, 'yi']
            e = self.metadata.loc[m, 'enhancement']

            # maxus value at the station
            if not pd.isnull(data_direction[m]):
                if self.config['station_peak'] is not None:
                    if m.upper() in self.config['station_peak']:
                        val_maxus = np.min(self.maxus[:, yi, xi] + e)

                else:
                    idx = int(np.ceil((data_direction[m] - self.nstep/2) /
                                      self.nstep) * self.nstep)
                    if idx == 360:
                        idx = 0  # special case when 360=0
                    ind = self.maxus_direction == idx

                    val_maxus = self.maxus[ind, yi, xi] + e

                # correct unreasonable values
                if val_maxus > 35:
                    val_maxus = 35
                if val_maxus < -35:
                    val_maxus = -35

                ma = np.abs(val_maxus)

                # Lapse all measurements to flat terrain (i.e. maxus = 0)
                if (ma > 21.3 and ma < 30.0):
                    expVal = 1.756507 - 0.1678945 * ma + \
                        0.01927844 * np.power(ma, 2) - \
                        0.0003651592 * np.power(ma, 3)
                elif (ma >= 30.0):
                    expVal = 4.21
                else:
                    expVal = 1.0 + 0.1031717 * (ma) - \
                        0.008003561 * np.power(ma, 2) + \
                        0.0003996581 * np.power(ma, 3)

                if val_maxus > 0:
                    flatwind.loc[m] = data_speed[m] * expVal
                else:
                    flatwind.loc[m] = data_speed[m] / expVal
            else:
                flatwind.loc[m] = np.NaN

        self.flatwind_point = flatwind

        # wind direction components at the station
        self.u_direction = np.sin(data_direction * np.pi/180)    # u
        self.v_direction = np.cos(data_direction * np.pi/180)    # v



import os
import shutil
from importlib.machinery import SourceFileLoader
from importlib.util import module_from_spec, spec_from_loader


def create_script_spec(script_name):

    # There are 3 places that the scripts can live
    # 1. In the dev environment
    # 2. In the docker image
    # 3. Installed through wheel
    SMRF_SCRIPT_HOME = [
        os.path.abspath(__file__ + '../../../../../') + '/scripts/',
        '/usr/local/bin/'
    ]

    found = False
    for sch in SMRF_SCRIPT_HOME:
        filename = os.path.join(sch, script_name)
        if os.path.exists(filename):
            SMRF_SCRIPT_HOME = filename
            found = True
            break

    # if it wasn't found then it should be installed
    if not found:
        SMRF_SCRIPT_HOME = shutil.which(script_name)

    return spec_from_loader(
        script_name,
        SourceFileLoader(
            script_name,
            SMRF_SCRIPT_HOME
        )
    )


def load_script_as_module(script_name):
    """
    Load given script file in the 'scripts' folder and load it as a module.

    Example:

    .. code-block:: Python

        from .script_test_helper import load_script_as_module

        script = load_script_as_module('script_name')

        script.method_from_script()

    Args:
        script_name: string
            Name of the script

    Returns:
        Script as module
    """
    spec = create_script_spec(script_name)
    script = module_from_spec(spec)
    spec.loader.exec_module(script)

    return script



import os

from smrf.tests.smrf_test_case import SMRFTestCase


class SMRFTestCaseLakes(SMRFTestCase):
    """
    Runs the short simulation over Lakes.
    """

    basin_dir = SMRFTestCase.test_dir.joinpath('basins', 'Lakes')
    config_file = os.path.join(basin_dir, SMRFTestCase.BASE_INI_FILE_NAME)
    gold_dir = basin_dir.joinpath('gold_hrrr')



import logging

import pandas as pd

from smrf.utils.utils import check_station_colocation


class InputCSV():

    DATA_TYPE = 'csv'

    def __init__(self, start_date, end_date, stations=None, config=None):

        self.start_date = start_date
        self.end_date = end_date
        self.stations = stations
        self.config = config
        self.time_zone = start_date.tzinfo

        self._logger = logging.getLogger(__name__)

        if self.stations is not None:
            self._logger.debug('Using only stations {0}'.format(
                ", ".join(self.stations)))

    def load(self):
        """
        Load the data from a csv file
        Fields that are operated on
        - metadata -> dictionary, one for each station,
        must have at least the following:
        primary_id, X, Y, elevation
        - csv data files -> dictionary, one for each time step,
        must have at least the following columns:
        date_time, column names matching metadata.primary_id
        """

        self._logger.info('Reading data coming from CSV files')

        variable_list = list(self.config.keys())
        variable_list.remove('stations')

        self._logger.debug('Reading {}...'.format(self.config['metadata']))
        metadata = pd.read_csv(
            self.config['metadata'],
            index_col='primary_id')
        # Ensure all stations are all caps.
        metadata.index = [s.upper() for s in metadata.index]
        self.metadata = metadata
        variable_list.remove('metadata')

        for variable in variable_list:
            filename = self.config[variable]

            self._logger.debug('Reading {}...'.format(filename))

            df = pd.read_csv(
                filename,
                index_col='date_time',
                parse_dates=[0])
            df = df.tz_localize(self.time_zone)
            df.columns = [s.upper() for s in df.columns]

            if self.stations is not None:
                df = df[df.columns[(df.columns).isin(self.stations)]]

            # Only get the desired dates
            df = df[self.start_date:self.end_date]

            if df.empty:
                raise Exception("No CSV data found for {0}"
                                "".format(variable))

            setattr(self, variable, df)

    def check_colocation(self):
        # Check all sections for stations that are colocated
        colocated = check_station_colocation(metadata=self.metadata)
        if colocated is not None:
            self._logger.error(
                "Stations are colocated: {}".format(','.join(colocated[0])))



"""
Created on March 14, 2017
Originally written by Scott Havens in 2015
@author: Micah Johnson

**Creating Custom NASDE Models**
--------------------------------
    When creating a new NASDE model make sure you adhere to the following:

    1. Add a new class to the nasde_models
        For example see :func:`~smrf.envphys.nasde_models.Susong1999`.

    #. Add the new model to the MODEL map in Snow :func:`~smrf.envphys.snow`

    #. Create a custom distribution function with a unique in
       :func:`~smrf.distribute.precipitation.distribute` to create the
       structure for the new model. For an example see
       :func:`~smrf.distribute.precipitation.distribute_for_susong1999`.

    #. Update documentation and run smrf!

"""

import numpy as np
from smrf.envphys.nasde_model import Marks2017, PiecewiseSusong1999, Susong1999
from types import MappingProxyType


class Snow:
    MODELS = MappingProxyType({
        "susong1999": Susong1999,
        "piecewise_susong1999": PiecewiseSusong1999,
        "marks2017": Marks2017,
    })

    @staticmethod
    def phase_and_density(temperature, precipitation, nasde_model):
        """
        Uses various new accumulated snow density models to estimate the snow
        density of precipitation that falls during sub-zero conditions.
        The models all are based on the dew point temperature and the amount of
        precipitation, All models used here must return a dictionary containing
        the keywords pcs and rho_s for percent snow and snow density
        respectively.

        Args:
            temperature: a single timestep of the distributed dew point
              temperature
            precipitation: a numpy array of the distributed precipitation
            nasde_model: string value set in the configuration file
              representing the method for estimating density of new snow that
              has just fallen.

        Returns:
            tuple:
            Returns a tuple containing the snow density field and the percent
            snow as determined by the NASDE model.

            - **snow_density** (*numpy.array*) - Snow density values in kg/m^3
            - **percent_snow** (*numpy.array*) -  Percent of the precipitation
                that is snow in values 0.0-1.0.

        """

        precipitation = np.array(precipitation)
        temperature = np.array(temperature)

        if nasde_model not in Snow.MODELS:
            raise ValueError(
                "{0} is not an implemented new accumulated snow density "
                "(NASDE) model! Check the config file under "
                "precipitation""".format(nasde_model))

        result = Snow.MODELS.get(nasde_model).run(temperature, precipitation)
        return result['rho_s'], result['pcs']



import logging.config
import os

import coloredlogs


class SMRFLogger():
    """Setup the root logger for SMRF to either the console or
    to a file
    """

    FMT = '%(levelname)s:%(name)s:%(message)s'

    def __init__(self, config):

        self.log_level = config.get('log_level', 'info').upper()
        self.log_file = config.get('log_file', None)

        # https://docs.python.org/3/library/logging.config.html#configuration-dictionary-schema
        log_config = {
            'version': 1,
            'formatters': {
                'standard': {
                    'format': self.FMT
                }
            },
            'handlers': {
                'default': {
                    'level': self.log_level,
                    'formatter': 'standard',
                    'class': 'logging.StreamHandler',
                    'stream': 'ext://sys.stdout',  # Default is stderr
                }
            },
            'loggers': {
                '': {  # root logger
                    'handlers': ['default'],
                    'level': self.log_level,
                    'propagate': False
                }
            }
        }

        if self.log_file is not None:
            os.makedirs(os.path.dirname(self.log_file), exist_ok=True)

            log_config['handlers']['log_file'] = {
                'level': self.log_level,
                'formatter': 'standard',
                'class': 'logging.FileHandler',
                'filename': self.log_file,
                'mode': 'a'
            }

            log_config['loggers']['']['handlers'] = ['log_file']

        logging.config.dictConfig(log_config)

        if self.log_file is None:
            coloredlogs.install(level=self.log_level, fmt=self.FMT)



import numpy as np

from .utils import check_temperature
from .piecewise_suosong_1999 import PiecewiseSusong1999


class Marks2017:
    """
    A new accumulated snow density model that accounts for compaction. The
    model builds upon :func:`~smrf.envphys.snow.piecewise_susong1999` by
    adding effects from compaction. Of four mechanisms for compaction, this
    model accounts for compaction by destructive metamorphism and overburden.
    These two processes are accounted for by calculating a proportionality
    using data from Kojima, Yosida and Mellor. The overburden is in part
    estimated using total storm deposition, where storms are defined in
    :func:`~smrf.envphys.storms.tracking_by_station`. Once this is determined
    the final snow density is applied through the entire storm only varying
    with hourly temperature.

    Snow Density:
        .. math::
            \\rho_{s} = \\rho_{ns} + (\\Delta \\rho_{c} + \\Delta
            \\rho_{m}) \\rho_{ns}

    Overburden Proportionality:
        .. math::
            \\Delta \\rho_{c} = 0.026 e^{-0.08 (T_{z} - T_{snow})}
            SWE*  e^{-21.0 \\rho_{ns}}

    Metamorphism Proportionality:
        .. math::
            \\Delta \\rho_{m} = 0.01 c_{11} e^{-0.04 (T_{z} - T_{snow})}

            c_{11} = c_min + (T_{z} - T_{snow}) C_{factor} + 1.0

    Constants:
        .. math::
            C_{factor} = 0.0013

            Tz = 0.0

            ex_{max} = 1.75

            exr = 0.75

            ex_{min} = 1.0

            c1r = 0.043

            c_{min} = 0.0067

            c_{fac} = 0.0013

            T_{min} = -10.0

            T_{max} = 0.0

            T_{z} = 0.0

            T_{r0} = 0.5

            P_{cr0} = 0.25

            P_{c0} = 0.75
    """

    # These should be used but unsure why not
    # ex_max = 1.75
    # exr = 0.75
    # ex_min = 1.0
    # #c1_min = 0.026
    # #c1_max = 0.069
    # c1r = 0.043
    # c_min = 0.0067
    # cfac = 0.0013
    T_MIN = -10.0
    T_MAX = 0.0
    TZ = 0.0
    # Tr0 = 0.5
    # Pcr0 = 0.25
    # Pc0 = 0.75

    RHO_WATER = 1000.0

    @staticmethod
    def run(tpp, pp):
        """
        Args:
            tpp: Numpy array of a single hour of temperature, use dew point if
                available [degrees C].
            pp: Numpy array representing the total amount of precipitation
                deposited during a storm in millimeters

        Returns:
            dictionary:
              - **rho_s** (*numpy.array*) - Density of the fresh snow in
                  kg/m^3.
              - **swe** (*numpy.array*) - Snow water equivalent.
              - **pcs** (*numpy.array*) - Percent of the precipitation that is
                  snow in values 0.0-1.0.
              - **rho_ns** (*numpy.array*) - Density of the uncompacted snow,
                  which is equivalent to the output from
                  :func:`~smrf.envphys.snow.piecewise_susong1999`.
              - **d_rho_c** (*numpy.array*) - Proportional coefficient for
                  compaction from overburden.
              - **d_rho_m** (*numpy.array*) - Proportional coefficient for
                  compaction from melt.
              - **rho_s** (*numpy.array*) - Final density of the snow [kg/m^3].
              - **rho** (*numpy.array*) - Density of the precipitation, which
                  continuously ranges from low density snow to pure liquid
                  water
                  (50-1000 kg/m^3).
              - **zs** (*numpy.array*) - Snow height added from the
                  precipitation

        """

        rho_ns = d_rho_m = d_rho_c = zs = rho_s = swe = pcs = np.zeros(
            tpp.shape
        )
        rho = np.ones(tpp.shape)

        if np.any(pp > 0):

            # check the precipitation temperature
            tpp, tsnow = check_temperature(
                tpp, t_max=Marks2017.T_MAX, t_min=Marks2017.T_MIN
            )

            # Calculate the percent snow and initial uncompacted density
            result = PiecewiseSusong1999.run(
                tpp, pp, t_max=Marks2017.T_MAX, t_min=Marks2017.T_MIN
            )
            pcs = result['pcs']
            rho_orig = result['rho_s']

            swe = pp * pcs

            # Calculate the density only where there is swe
            swe_ind = swe > 0.0
            if np.any(swe_ind):

                s_pcs = pcs[swe_ind]
                s_pp = pp[swe_ind]
                s_swe = swe[swe_ind]
                # transforms to a 1D array, will put back later
                s_tsnow = tsnow[swe_ind]

                # transforms to a 1D array, will put back later
                s_rho_ns = rho_orig[swe_ind]

                # Convert to a percentage of water
                s_rho_ns = s_rho_ns/Marks2017.RHO_WATER

                # proportional total storm mass compaction
                s_d_rho_c = (0.026 * np.exp(-0.08 * (Marks2017.TZ - s_tsnow))
                             * s_swe * np.exp(-21.0 * s_rho_ns))

                ind = (s_rho_ns * Marks2017.RHO_WATER) >= 100.0
                c11 = np.ones(s_rho_ns.shape)

                # c11[ind] = (c_min + ((Tz - s_tsnow[ind]) * cfac)) + 1.0

                c11[ind] = np.exp(-0.046*(s_rho_ns[ind]*1000.0-100.0))

                s_d_rho_m = 0.01 * c11 * np.exp(
                    -0.04 * (Marks2017.TZ - s_tsnow)
                )

                # Snow density, depth & combined liquid and snow density
                s_rho_s = s_rho_ns + ((s_d_rho_c + s_d_rho_m) * s_rho_ns)

                s_zs = s_swe / s_rho_s

                # a mixture of snow and liquid
                s_rho = s_rho_s.copy()
                mix = (s_swe < s_pp) & (s_pcs > 0.0)
                if np.any(mix):
                    s_rho[mix] = \
                        (s_pcs[mix] * s_rho_s[mix]) + (1.0 - s_pcs[mix])

                s_rho[s_rho > 1.0] = 1.0

                # put the values back into the larger array
                rho_ns[swe_ind] = s_rho_ns
                d_rho_m[swe_ind] = s_d_rho_m
                d_rho_c[swe_ind] = s_d_rho_c
                zs[swe_ind] = s_zs
                rho_s[swe_ind] = s_rho_s
                rho[swe_ind] = s_rho
                pcs[swe_ind] = s_pcs

        # convert densities from proportions, to kg/m^3 or mm/m^2
        rho_ns = rho_ns * Marks2017.RHO_WATER
        rho_s = rho_s * Marks2017.RHO_WATER
        rho = rho * Marks2017.RHO_WATER

        return {
            'swe': swe,
            'pcs': pcs,
            'rho_ns': rho_ns,
            'd_rho_c': d_rho_c,
            'd_rho_m': d_rho_m,
            'rho_s': rho_s,
            'rho': rho,
            'zs': zs
        }



from __future__ import print_function

import os
# import matplotlib.pyplot as plt
# import progressbar
from datetime import datetime

import netCDF4 as nc
import numpy as np

from smrf.utils.wind import wind_c


class wind_model():
    """

    Estimating wind speed and direction is complex terrain can be difficult due
    to the interaction of the local topography with the wind. The methods
    described here follow the work developed by Winstral and Marks (2002) and
    Winstral et al. (2009) :cite:`Winstral&Marks:2002` :cite:`Winstral&al:2009`
    which parameterizes the terrain based on the upwind direction. The
    underlying method calulates the maximum upwind slope (maxus) within a
    search distance to determine if a cell is sheltered or exposed.

    The azimuth **A** is the direction of the prevailing wind for which the
    maxus value will be calculated within a maximum search distance **dmax**.
    The maxus (**Sx**) parameter can then be estimated as the maximum value of
    the slope from the cell of interest to all of the grid cells along the
    search vector. The efficiency in selection of the maximum value can be
    increased by using the techniques from the horizon functio which calculates
    the horizon for each pixel. Therefore, less calculations can be performed.
    Negative **Sx** values indicate an exposed pixel location (shelter pixel
    was lower) and positive **Sx** values indicate a sheltered pixel (shelter
    pixel was higher).

    After all the upwind direction are calculated, the average **Sx** over a
    window is calculated. The average **Sx** accounts for larger lanscape
    obsticles that may be adjacent to the upwind direction and affect the flow.
    A window size in degrees takes the average of all **Sx**.

    Args:
        x: array of x locations
        y: array of y locations
        dem: matrix of the dem elevation values
        nthread: number of threads to use for maxus calculation

    """

    def __init__(self, x, y, dem, nthreads=1):

        self.x = x
        self.y = y
        self.dem = dem
        self.nx = len(x)
        self.ny = len(y)
        self.ngrid = self.ny * self.nx
        self.nthreads = nthreads

        self.dx = np.abs(x[1] - x[0])
        self.dy = np.abs(y[1] - y[0])

        X, Y = np.meshgrid(np.arange(0, self.nx), np.arange(0, self.ny))
        self.X = X
        self.Y = Y
        self.shape = X.shape

    def maxus(self, dmax, inc=5, inst=2, out_file='smrf_maxus.nc'):
        """
        Calculate the maxus values

        Args:
            dmax: length of outlying upwind search vector (meters)
            inc: increment between direction calculations (degrees)
            inst: Anemometer height (meters)
            out_file: NetCDF file for output results

        Returns:
            None, outputs maxus array straight to file
        """

        if (dmax % self.dx != 0):
            raise ValueError('dmax must divide evenly into the DEM')

        self.dmax = dmax
        self.inc = inc
        self.inst_hgt = inst

        # All angles that model will consider.
        swa = np.arange(0, 360, inc)
        self.directions = swa

        # initialize the output file
        self.out_file = out_file
        self.type = 'maxus'
        ex_att = {}
        ex_att['dmax'] = dmax
        # initialize output
        self.output_init(self.type, out_file, ex_att=ex_att)

        # run model over range in wind directions
        for i, angle in enumerate(swa):

            self.maxus_val = self.maxus_angle(angle, self.dmax)
            self.output(self.type, i)

    def tbreak(self, dmax, sepdist, inc=5, inst=2, out_file='smrf_tbreak.nc'):
        """
        Calculate the topobreak values

        Args:
            dmax: length of outlying upwind search vector (meters)
            sepdist: length of local max upwind slope search vector (meters)
            angle: middle upwind direction around which to run model (degrees)
            inc: increment between direction calculations (degrees)
            inst: Anemometer height (meters)
            out_file: NetCDF file for output results

        Returns:
            None, outputs maxus array straight to file

        """

        if (sepdist % self.dx != 0) | (dmax % self.dx != 0):
            raise ValueError(
                'sepdist and dmax must divide evenly into the DEM')

        self.dmax = dmax
        self.sepdist = sepdist
        self.inc = inc
        self.inst_hgt = inst

        # All angles that model will consider.
        swa = np.arange(0, 360, inc)
        self.directions = swa

        # initialize the output file
        self.out_file = out_file
        self.type = 'tbreak'
        # extra attributes
        ex_att = {}
        ex_att['dmax'] = dmax
        ex_att['sepdist'] = sepdist
        # initialize output
        self.output_init(self.type, out_file, ex_att=ex_att)

        # run model over range in wind directions
        for i, angle in enumerate(swa):

            # calculate the maxus value
            maxus_outlying = self.maxus_angle(angle, self.dmax)

            # calculate the local maxus value
            maxus_local = self.maxus_angle(angle, self.sepdist)

            self.maxus_val = maxus_local - maxus_outlying

            self.output(self.type, i)

    def maxus_angle(self, angle, dmax):
        """
        Calculate the maxus for a single direction for a search distance dmax

        Note:
            This will produce different results than the original maxus
            program. The differences are due to:

            1. Using dtype=double for the elevations
            2. Using different type of search method to find the endpoints.

            However, if the elevations are rounded to integers, the cardinal
            directions will reproduce the original results.

        Args:
            angle: middle upwind direction around which to run model (degrees)
            dmax: length of outlying upwind search vector (meters)

        Returns:
            maxus: array of maximum upwind slope values within dmax

        """

        print("Calculating maxus for direction: {}".format(angle))

        angle *= np.pi / 180

        # calculate the endpoints
        # accually use the distances to ensure that we are searching far enough
        Xi = self.X*self.dx + dmax * np.cos(angle-np.pi/2)
        Yi = self.Y*self.dy + dmax * np.sin(angle-np.pi/2)

        self.Xi = np.floor(Xi/self.dx + 0.5)
        self.Yi = np.floor(Yi/self.dy + 0.5)

        # underlying C code similar to Adams
        maxus = wind_c.call_maxus(self.x, self.y, self.dem, self.X, self.Y,
                                  self.Xi, self.Yi, self.inst_hgt,
                                  self.nthreads)

#         # my interpretation of the calculations in Python form
#         maxus = np.zeros((self.ngrid,))
#         pbar = progressbar.ProgressBar(max_value=self.ngrid)
#         j = 0
#         for index in range(5000, self.ngrid):
#             maxus[index] = self.find_maxus(index)
#             j += 1
#             pbar.update(j)
#             if j > 4999:
#                 break
#         pbar.finish()
#         maxus = np.reshape(maxus, self.shape)

        # correct for values that are their own horizon
        maxus[maxus <= -89.0] = 0

        return maxus

    def windower(self, maxus_file, window_width, wtype):
        """
        Take the maxus output and average over the window width

        Args:
            maxus_file: location of the previously calculated maxus values
            window_width: window width about the wind direction
            wtype: type of wind calculation 'maxus' or 'tbreak'

        Return:
            New file containing the windowed values
        """

        # open the previous file and get the directions
        n = nc.Dataset(maxus_file, 'r')
        directions = n.variables['direction'][:]
        self.directions = directions

        # create a new file based on the old file
        name = os.path.splitext(maxus_file)
        out_file = "%s_%iwindow.nc" % (name[0], window_width)
        self.output_init(wtype, out_file)
        self.out_file = out_file

        # determine which directions are required for each single direction
        window_width /= 2
        inc = np.mean(np.diff(directions), dtype=int)

        for i, d in enumerate(directions):

            print("Windowing direction {}".format(d))

            # determine which directions to include
            window_start = d - window_width
            window_end = d + window_width

            # to ensure that it contains the end points
            sl = np.arange(window_start, window_end+1, inc)

            # correct for edge effects
            sl[sl < 0] = sl[sl < 0] + 360
            sl[sl > 360] = sl[sl > 360] - 360

            # determine the indicies to the input file
            idx = self.ismember(directions, sl)

            # grab all the data for all directions and average
            self.maxus_val = np.mean(n.variables[wtype][idx, :], axis=0)

            # put it into the output file
            self.output(wtype, i)

        n.close()

    def ismember(self, a, b):
        bind = {}
        for i, elt in enumerate(b):
            if elt not in bind:
                bind[elt] = True
        return [bind.get(itm, False) for itm in a]

    def find_maxus(self, index):
        """
        Calculate the maxus given the start and end point

        Args:
            index: index to a point in the array

        Returns:
            maxus value for the point
        """

        start_point = np.unravel_index(index, self.shape)

        # determine the points along the endpoint line
        end_point = (self.Yi[start_point], self.Xi[start_point])
        p = self.bresenham(start_point, end_point)

        # ensure the cases where it's on the edge
        p = np.delete(p, np.where(p[:, 0] < 0)[0], axis=0)
        p = np.delete(p, np.where(p[:, 1] < 0)[0], axis=0)
        p = np.delete(p, np.where(p[:, 0] > self.ny)[0], axis=0)
        p = np.delete(p, np.where(p[:, 1] > self.nx)[0], axis=0)

        # determine the relative heights along the path
        h = self.dem[p[:, 0], p[:, 1]]  # - (self.inst_hgt + self.dem[index])

        # find the horizon for each pixel along the path
        hord = self.hord(self.x[p[:, 1]], self.y[p[:, 0]], h)

        # calculate the angle to that point
        pt = p[hord[0], :]   # point that was found for horizon
        d = np.sqrt(np.power(self.x[pt[1]] - self.x[start_point[1]], 2) +
                    np.power(self.y[pt[0]] - self.y[start_point[0]], 2))

        slope = (h[hord[0]] - (h[0] + self.inst_hgt)) / d
        maxus = np.arctan(slope) * 180 / np.pi

        return maxus

    def bresenham(self, start, end):
        """
        Python implementation of the Bresenham algorthim to find
        all the pixels that a line between start and end interscet

        Args:
            start: list of start point
            end: list of end point

        Returns:
            Array path of all points between start and end
        """
#         start = list(start)
#         end = list(end)
        path = []

        x0 = start[0]
        y0 = start[1]
        x1 = end[0]
        y1 = end[1]

        steep = abs(y1 - y0) > abs(x1 - x0)
        backward = x0 > x1

        if steep:
            x0, y0 = y0, x0
            x1, y1 = y1, x1
        if backward:
            x0, x1 = x1, x0
            y0, y1 = y1, y0

        dx = x1 - x0
        dy = abs(y1 - y0)
        error = dx / 2
        y = y0

        if y0 < y1:
            ystep = 1
        else:
            ystep = -1

        for x in range(x0, x1+1):
            if steep:
                path.append((y, x))
            else:
                path.append((x, y))

            error -= dy

            if error <= 0:
                y += ystep
                error += dx

        if backward:
            path.reverse()

        return np.array(path)

    def hord(self, x, y, z):
        '''
        Calculate the horizon pixel for all z
        This mimics the simple algorthim from Dozier 1981 but
        was adapated for use in finding the maximum upwind slope

        Works backwards from the end but looks forwards for
        the horizon

        Args:
            x: x locations for the points
            y: y locations for the points
            z: elevations for the points

        Returns:
            array of the horizon index for each point

        '''

        N = len(z)  # number of points to look at
    #     offset = 1      # offset from current point to start looking

        # preallocate the h array
        h = np.zeros(N, dtype=int)
        h[N-1] = N-1
        i = N - 2

        # work backwarks from the end for the pixels
        while i >= 0:
            h[i] = i
            j = i + 1   # looking forward
            found = False

            while not found:

                d_i = np.sqrt(np.power(x[i] - x[j], 2) +
                              np.power(y[i] - y[j], 2))
                d_h = np.sqrt(np.power(x[i] - x[h[j]], 2) +
                              np.power(y[i] - y[h[j]], 2))

                pt_i = self._slope(0, z[i]+self.inst_hgt, d_i, z[j])
                pt_h = self._slope(0, z[i]+self.inst_hgt, d_h, z[h[j]])

                if (pt_i < pt_h):
                    if (j == N-1):
                        found = True
                        h[i] = j
                    else:
                        j = h[j]
                else:
                    found = True
                    if (pt_i > pt_h):
                        h[i] = j
                    else:
                        h[i] = h[j]

            i -= 1

        return h

    def _slope(self, xi, zi, xj, zj):
        '''
        Slope between the two points
        '''

        return (zj - zi) / (xj - float(xi))

    def output_init(self, ptype, filename, ex_att=None):
        """
        Initialize a NetCDF file for outputing the maxus values or tbreak

        Args:
            ptype: type of calculation that will be saved, either 'maxus' or
                'tbreak'
            filename: filename to save the output into
            ex_att:   extra attributes to add
        """

        if ptype == 'maxus':
            var = 'maxus'
            desc = 'Maximum upwind slope'

        elif ptype == 'tbreak':
            var = 'tbreak'
            desc = 'tbreak'

        else:
            raise ValueError('''Could not determine what to output, check type
                            value (maxus or tbreak)''')

        dimensions = ('Direction', 'y', 'x')

        s = nc.Dataset(filename, 'w', 'NETCDF4')

        s.createDimension(dimensions[0], len(self.directions))
        s.createDimension(dimensions[1], self.ny)
        s.createDimension(dimensions[2], self.nx)

        # create the variables
        s.createVariable('direction', 'i', dimensions[0])
        s.createVariable('y', 'f', dimensions[1])
        s.createVariable('x', 'f', dimensions[2])
        s.createVariable(var, 'f', dimensions)

        # define some attributes
        setattr(s.variables['y'], 'units', 'meters')
        setattr(s.variables['y'], 'description', 'UTM, north south')
        setattr(s.variables['x'], 'units', 'meters')
        setattr(s.variables['x'], 'description', 'UTM, east west')
        setattr(s.variables['direction'], 'units', 'bearing')
        setattr(s.variables['direction'], 'description',
                'Wind direction from North')
        setattr(s.variables[var], 'units', 'angle')
        setattr(s.variables[var], 'description', desc)
        setattr(s, 'dateCreated', datetime.now().isoformat())

        # set attributes
        if ex_att is not None:
            for key, value in ex_att.items():
                setattr(s, key, value)

        s.variables['y'][:] = self.y
        s.variables['x'][:] = self.x

    def output(self, ptype, index):
        """
        Output the data into the out file that has previously been initialized.

        Args:
            ptype: type of calculation that will be saved, either 'maxus'
                or 'tbreak'
            index: index into the file for where to place the output
        """

        s = nc.Dataset(self.out_file, 'r+')
        s.variables['direction'][:] = self.directions
        s.variables[ptype][index, :] = self.maxus_val
        s.close()



from smrf.envphys.nasde_model import Marks2017
import unittest
import numpy as np


class TestMarks2017(unittest.TestCase):
    TEMPERATURES = np.array([-10, -5, 0, 1])
    PRECIPITATION = np.array([100, 100, 100, 100])

    def test_returns_precipitation(self):
        results = Marks2017.run(self.TEMPERATURES, self.PRECIPITATION)

        np.testing.assert_almost_equal(
            results['rho_s'], [79, 115, 247, 0], decimal=0
        )
        assert 'pcs' in results



# flake8: noqa
import sys

PY3 = sys.version_info[0] >= 3

if PY3:  # pragma: no cover
    basestring = str
    unicode_type = str
    bytes_type = bytes

    def iteritems(d):
        return iter(d.items())

    def itervalues(d):
        return iter(d.values())

    pyrange = range
    pyzip = zip
    import builtins
    from configparser import SafeConfigParser
    from functools import reduce as pyreduce
else:  # pragma: no cover
    # Python 2
    basestring = basestring
    unicode_type = unicode
    bytes_type = str

    def iteritems(d):
        return d.iteritems()

    def itervalues(d):
        return d.itervalues()

    pyrange = xrange
    from itertools import imap as pymap
    from itertools import izip as pyzip
    pyreduce = reduce
    import __builtin__ as builtins
    from ConfigParser import SafeConfigParser


try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict



from smrf.envphys.nasde_model import PiecewiseSusong1999
import unittest
import numpy as np


class TestPiecewiseSusong1999(unittest.TestCase):
    TEMPERATURES = np.random.uniform(low=0, high=3, size=12)
    PRECIPITATION = np.random.uniform(
        low=0.5, high=30, size=TEMPERATURES.size
    )

    def test_returns_precipitation(self):
        results = PiecewiseSusong1999.run(
            self.TEMPERATURES, self.PRECIPITATION
        )

        assert 'rho_s' in results
        assert 'pcs' in results



import logging

import netCDF4 as nc
import numpy as np
import pandas as pd

from smrf.utils.utils import apply_utm


def metadata_name_from_index(index):
    return 'grid_y{:d}_x{:d}'.format(index[0], index[1])


class InputNetcdf():

    DATA_TYPE = 'netcdf'

    VARIABLES = [
        'air_temp',
        'vapor_pressure',
        'precip',
        'wind_speed',
        'wind_direction',
        'cloud_factor'
    ]

    def __init__(self, start_date, end_date, bbox=None,
                 topo=None, config=None):

        self.start_date = start_date
        self.end_date = end_date
        self.topo = topo
        self.bbox = bbox
        self.config = config
        self.time_zone = start_date.tzinfo

        if topo is None:
            raise Exception('Must supply topo to InputWRF')

        if bbox is None:
            raise Exception('Must supply bbox to InputWRF')

        self.variables = list(self.config.keys())
        self.variables.remove('data_type')
        self.variables.remove('netcdf_file')

        self._logger = logging.getLogger(__name__)

    def load(self):
        """
        Load the data from a generic netcdf file

        Args:
            lat: latitude field in file, 1D array
            lon: longitude field in file, 1D array
            elev: elevation field in file, 2D array
            variable: variable name in file, 3D array
        """

        self._logger.info('Reading data coming from netcdf: {}'.format(
            self.config['netcdf_file'])
        )

        f = nc.Dataset(self.config['netcdf_file'], 'r')

        # GET THE LAT, LON, ELEV FROM THE FILE
        mlat = f.variables['lat'][:]
        mlon = f.variables['lon'][:]
        mhgt = f.variables['elev'][:]

        if mlat.ndim != 2 & mlon.ndim != 2:
            [mlon, mlat] = np.meshgrid(mlon, mlat)

        # get the values that are in the modeling domain
        ind = (mlat >= self.bbox[1]) & \
            (mlat <= self.bbox[3]) & \
            (mlon >= self.bbox[0]) & \
            (mlon <= self.bbox[2])

        mlat = mlat[ind]
        mlon = mlon[ind]
        mhgt = mhgt[ind]

        # GET THE METADATA
        # create some fake station names based on the index
        a = np.argwhere(ind)
        primary_id = [metadata_name_from_index(i) for i in a]
        self._logger.debug('{} grid cells within model domain'.format(len(a)))

        # create a metadata dataframe to store all the grid info
        metadata = pd.DataFrame(index=primary_id,
                                columns=('X', 'Y', 'latitude',
                                         'longitude', 'elevation'))

        metadata['latitude'] = mlat.flatten()
        metadata['longitude'] = mlon.flatten()
        metadata['elevation'] = mhgt.flatten()
        metadata = metadata.apply(apply_utm,
                                  args=(self.topo.zone_number,),
                                  axis=1)

        self.metadata = metadata

        # GET THE TIMES
        t = f.variables['time']
        time = nc.num2date(t[:].astype(int), t.getncattr(
            'units'), t.getncattr('calendar'))
        # Drop milliseconds and prepare to use as pandas DataFrame index
        time = pd.DatetimeIndex(
            [str(tm.replace(microsecond=0)) for tm in time], tz=self.time_zone
        )

        # GET THE DATA, ONE AT A TIME
        for variable in self.VARIABLES:
            v_file = self.config[variable]
            self._logger.debug(
                'Loading variable {} from netcdf field {}'.format(
                    variable, v_file))

            df = pd.DataFrame(index=time, columns=primary_id)
            for i in a:
                g = metadata_name_from_index(i)
                df[g] = f.variables[v_file][:, i[0], i[1]]

            # deal with any fillValues
            try:
                fv = f.variables[v_file].getncattr('_FillValue')
                df.replace(fv, np.nan, inplace=True)
            except Exception:
                pass

            # Set variable and subset by start and end time
            setattr(self, variable, df)



from smrf.tests.recipes.base_recipe import BaseRecipes
from smrf.tests.smrf_test_case import SMRFTestCase
from smrf.tests.smrf_test_case_lakes import SMRFTestCaseLakes


class TestWindNinjaRecipes(SMRFTestCaseLakes, BaseRecipes):
    """Test the wind ninja recipes
    """

    def test_wind_ninja_recipe(self):
        """Test the wind ninja recipe
        """

        # get the master config list
        master_config = self.master_config(self.base_config)

        # make changes
        master_config['wind'] = [
            'wind_model',
            'wind_ninja_dir',
            'wind_ninja_dxdy',
            'wind_ninja_pref',
            'wind_ninja_tz',
            'wind_ninja_roughness',
            'wind_ninja_height',
            'min',
            'max'
        ]

        self.check_config_section(self.base_config, master_config, 'wind')


class TestWinstralWindRecipes(SMRFTestCase, BaseRecipes):
    """Test the winstral recipes
    """

    def test_winstral_wind_recipe(self):
        """Test the winstral wind recipe
        """

        # get the master config list
        master_config = self.master_config(self.base_config)

        # make changes
        master_config['wind'] = [
            'maxus_netcdf',
            'reduction_factor',
            'wind_model',
            'distribution',
            'detrend',
            'detrend_slope',
            'stations',
            'max',
            'min',
            'idw_power',
            'station_peak',
            'station_default',
            'veg_default',
            'veg_41',
            'veg_42',
            'veg_43',
            'veg_3011',
            'veg_3061'
        ]

        self.check_config_section(self.base_config, master_config, 'wind')


class TestInterpWindRecipes(SMRFTestCase, BaseRecipes):
    """Test the interp recipes
    """

    def test_interp_wind_recipe(self):
        """Test the intper wind recipe
        """
        config = self.base_config_copy()

        adj_config = {
            'wind': {
                'distribution': 'grid',
                'wind_model': 'interp'
            }
        }

        config.raw_cfg.update(adj_config)
        config = self.cast_recipes(config)

        # get the master config list
        master_config = self.master_config(config)

        # make changes
        master_config['wind'] = [
            'wind_model',
            'distribution',
            'detrend',
            'detrend_slope',
            'max',
            'min',
            'grid_mask',
            'grid_method',
            'grid_local'
        ]

        self.check_config_section(config, master_config, 'wind')



import numpy as np

from smrf.distribute import image_data


class ts(image_data.image_data):
    """
    The :mod:`~smrf.distribute.soil_temp.ts` class allows for variable specific
    distributions that go beyond the base class.

    Soil temperature is simply set to a constant value during initialization.
    If soil temperature measurements are available, the values can be
    distributed using the distribution methods.

    Args:
        soilConfig: The [soil] section of the configuration file

    Attributes:
        config: configuration from [soil] section
        soil_temp: numpy array of the soil temperature
        stations: stations to be used in alphabetical order

    """

    variable = 'soil_temp'

    # these are variables that can be output
    OUTPUT_VARIABLES = {
        'soil_temp': {
            'units': 'degree_Celcius',
            'standard_name': 'soil_temperature',
            'long_name': 'Soil temperature'
        }
    }

    # these are variables that are operate at the end only and do not need to
    # be written during main distribute loop
    post_process_variables = {}

    def __init__(self, soilConfig):

        # extend the base class
        image_data.image_data.__init__(self, self.variable)

        self.config = soilConfig

        self._logger.debug('Created distribute.soil_temp')

    def initialize(self, topo, data, date_time=None):
        """
        Initialize the distribution and set the soil temperature to a constant
        value based on the configuration file.

        Args:
            topo: :mod:`smrf.data.loadTopo.Topo` instance contain topographic
                data and infomation
            metadata: data Pandas dataframe containing the station data,
                from :mod:`smrf.data.loadData` or :mod:`smrf.data.loadGrid`

        """

        self._logger.debug('Initializing distribute.soil_temp')
        self.date_time = date_time
#         self._initialize(topo, metadata)
        self.soil_temp = float(self.config['temp']) * np.ones(topo.dem.shape)

    def distribute(self):
        """
        No distribution is performed on soil temperature at the moment, method
        simply passes.

        Args:
            None
        """

#         self._logger.debug('Distributing soil temp')
        pass



import logging

import netCDF4 as nc
import numpy as np
import pandas as pd
import pytz

from smrf.data.netcdf import metadata_name_from_index
from smrf.envphys.vapor_pressure import satvp
from smrf.utils.utils import apply_utm


class InputWRF():

    DATA_TYPE = 'wrf'

    BASE_VARIABLES = {
        'air_temp': 'T2',
        'dew_point': 'DWPT',
        'precip': 'RAINNC'
    }

    WRF_VARIABLES = {
        'air_temp': 'T2',
        'dew_point': 'DWPT',
        'u10': 'UGRD',
        'v10': 'VGRD',
        'cloud_factor': 'CLDFRA',
        'precip': 'RAINNC'
    }

    def __init__(self, start_date, end_date, bbox=None,
                 topo=None, config=None):

        self.start_date = start_date
        self.end_date = end_date
        self.topo = topo
        self.bbox = bbox
        self.config = config
        self.time_zone = start_date.tzinfo

        if topo is None:
            raise Exception('Must supply topo to InputWRF')

        if bbox is None:
            raise Exception('Must supply bbox to InputWRF')

        self._logger = logging.getLogger(__name__)

    def get_file_times(self):
        """Read the times from the WRF file and attempt to
        interpret the format.

        Raises:
            Exception: could not parse the times

        Returns:
            list: list of date times
        """

        t = self.file.variables['Times']
        t.set_auto_maskandscale(False)
        try:
            times = [('').join(v) for v in t]
        except TypeError:
            times = []
            for v in t:
                times.append(''.join([s.decode('utf-8') for s in v]))
        except Exception:
            raise Exception('Could not convert WRF times to readable format')

        times = [v.replace('_', ' ') for v in times]  # remove the underscore
        in_utc = str(self.time_zone).lower() == str(pytz.UTC).lower()
        time = pd.to_datetime(times, utc=in_utc)

        return time

    def get_metadata(self):

        # get the values that are in the modeling domain
        ind = (self.file.variables['XLAT'] >= self.bbox[1]) & \
            (self.file.variables['XLAT'] <= self.bbox[3]) & \
            (self.file.variables['XLONG'] >= self.bbox[0]) & \
            (self.file.variables['XLONG'] <= self.bbox[2])

        mlat = self.file.variables['XLAT'][:][ind]
        mlon = self.file.variables['XLONG'][:][ind]
        mhgt = self.file.variables['HGT'][:][ind]

        # GET THE METADATA
        # create some fake station names based on the index
        self.station_index = np.argwhere(ind)
        self.primary_id = [metadata_name_from_index(
            i) for i in self.station_index]
        self._logger.debug(
            '{} grid cells within model domain'.format(
                len(self.station_index)))

        # create a metadata dataframe to store all the grid info
        metadata = pd.DataFrame(index=self.primary_id,
                                columns=('utm_x', 'utm_y', 'latitude',
                                         'longitude', 'elevation'))

        metadata['latitude'] = mlat.flatten()
        metadata['longitude'] = mlon.flatten()
        metadata['elevation'] = mhgt.flatten()
        metadata = metadata.apply(apply_utm,
                                  args=(self.topo.zone_number,),
                                  axis=1)

        self.metadata = metadata

    def load(self):
        """
        Load the data from a netcdf file. This was setup to work with a WRF
        output file, i.e. wrf_out so it's going to look for the following
        variables:
        - Times
        - XLAT
        - XLONG
        - HGT
        - T2
        - DWPT
        - GLW
        - RAINNC
        - CLDFRA
        - UGRD
        - VGRD

        Each cell will be identified by grid_IX_IY

        """

        self._logger.info('Reading data coming from WRF output: {}'.format(
            self.config['wrf_file']
        ))
        self.file = nc.Dataset(self.config['wrf_file'])

        self.get_metadata()

        # GET THE TIMES
        time = self.get_file_times()
        time_ind = (time >= self.start_date) & (time <= self.end_date)
        time = time[time_ind]

        # GET THE DATA, ONE AT A TIME
        for variable, wrf_variable in self.BASE_VARIABLES.items():

            self._logger.debug(
                'Loading variable {} from WRF field {}'.format(
                    variable, wrf_variable))

            df = pd.DataFrame(index=time, columns=self.primary_id)
            for i in self.station_index:
                g = metadata_name_from_index(i)
                df[g] = self.file.variables[wrf_variable][time_ind, i[0], i[1]]

            # Set variable and subset by start and end time
            setattr(self, variable, df)

        # post process the loaded data
        self.air_temp = self.air_temp - 273.15

        self._logger.debug('Calculating vapor_pressure')
        vp = satvp(self.dew_point.values)
        self.vapor_pressure = pd.DataFrame(
            vp, index=time, columns=self.primary_id)

        self._logger.debug('Loading cloud_factor')
        self.cloud_factor = pd.DataFrame(index=time, columns=self.primary_id)
        cf = 1 - np.mean(self.file.variables['CLDFRA'][time_ind, :], axis=1)
        for i in self.station_index:
            g = metadata_name_from_index(i)
            v = cf[:, i[0], i[1]]
            self.cloud_factor[g] = v

        self._logger.debug('Loading wind_speed and wind_direction')
        self.wind_speed = pd.DataFrame(index=time, columns=self.primary_id)
        self.wind_direction = pd.DataFrame(index=time, columns=self.primary_id)

        u10 = self.file.variables['UGRD'][time_ind, :]
        v10 = self.file.variables['VGRD'][time_ind, :]

        # calculate the wind speed
        s = np.sqrt(u10**2 + v10**2)

        # calculate the wind direction
        d = np.degrees(np.arctan2(v10, u10))
        ind = d < 0
        d[ind] = d[ind] + 360

        for i in self.station_index:
            g = metadata_name_from_index(i)
            self.wind_speed[g] = s[:, i[0], i[1]]
            self.wind_direction[g] = d[:, i[0], i[1]]

        self._logger.debug('Loading precip')
        self.precip = self.precip.diff(axis=0)
        self.precip.iloc[0, :] = 0



"""
2016-02-22 Scott Havens

Distributed forcing data over a grid using detrended kriging
"""

import logging

import numpy as np
import pandas as pd

from . import detrended_kriging


class DK:
    """
    Detrended kriging class
    """

    def __init__(self, mx, my, mz, GridX, GridY, GridZ, config):
        """
        Args:
            mx: x locations for the points
            my: y locations for the points
            mz: z locations for the points
            GridX: x locations in grid to interpolate over
            GridY: y locations in grid to interpolate over
            GridZ: z locations in grid to interpolate over
        """

        # measurement point locations
        self.mx = mx
        self.my = my
        self.mz = mz
        self.nsta = len(mx)

        # grid information
        self.GridX = GridX
        self.GridY = GridY
        self.GridZ = GridZ
        self.ngrid = GridX.shape[0] * GridX.shape[1]

        # data information
        self.data = None
        self.nan_val = []

        # initialize some things
        self.weights = []
        self.dgrid = []
        self.ad = []

        self.config = config

        # calculate the distances
#         self.calculateDistances()

        # calculate the weights
#         self.calculateWeights()

        self._logger = logging.getLogger(__name__)

    def calculate(self, data):
        """
        Calcluate the deternded kriging for the data and config

        Arg:
            data: numpy array same length as m*
            config: configuration for dk

        Returns:
            v: returns the distributed and calculated value
        """

        nan_val = pd.isnull(data)

        # only calcualte if the stations involved have changed
        # if np.sum(np.array_equal(nan_val,self.nan_val)) != len(self.mx):
        if not np.array_equal(nan_val, self.nan_val):

            self.nan_val = nan_val
            nsta = np.sum(~nan_val)
            self._logger.debug('''Recalculating detrended kriging weights
                                for {} stations ...'''.format(nsta))

            self.calculateWeights()

        # now calculate the trend and the residuals
        self.detrendData(data)

        # distribute the risduals
        r = np.nansum(self.weights * self.residuals, 2)

        # retrend the residuals
        v = self.retrendData(r)

        return v

    def calculateWeights(self):
        """
        Calculate the weights given those stations with nan values for data
        """

        nsta = np.sum(~self.nan_val)
        mx = self.mx[~self.nan_val]
        my = self.my[~self.nan_val]
        mz = self.mz[~self.nan_val]

        # calculate the distances between stations
        ad = np.zeros((nsta, nsta))
        for i in range(nsta):
            ad[i, i] = 0
            for j in range(i+1, nsta):
                ad[i, j] = np.sqrt((mx[i] - mx[j])**2 + (my[i] - my[j])**2)
                ad[j, i] = np.sqrt((mx[i] - mx[j])**2 + (my[i] - my[j])**2)

        self.ad = ad

        # calculate the distances from the grid to the station
        dgrid = np.zeros((self.ngrid, nsta))
        Xa = self.GridX.ravel()
        Ya = self.GridY.ravel()
        for i in range(nsta):
            dgrid[:, i] = np.sqrt((Xa - mx[i])**2 + (Ya - my[i])**2)

        self.dgrid = dgrid

        # calculate the weights
        wg = np.zeros_like(dgrid)
        detrended_kriging.call_grid(self.ad, self.dgrid, mz.astype(np.double),
                                    wg, self.config['dk_ncores'])

        # reshape the weights
        self.weights = np.zeros((self.GridX.shape[0],
                                 self.GridX.shape[1],
                                 nsta))
        for v in range(nsta):
            self.weights[:, :, v] = wg[:, v].reshape(self.GridX.shape)

    def detrendData(self, data):
        """
        Detrend the data in val using the heights zmeas
        data    - is the same size at mx,my
        flag     - 1 for positive, -1 for negative, 0 for any trend imposed
        """

        # calculate the trend on any real data
        pv = np.polyfit(self.mz[~self.nan_val], data[~self.nan_val], 1)

        # apply trend constraints
        if self.config['detrend_slope'] == 1 and pv[0] < 0:
            pv = np.array([0, 0])
        elif (self.config['detrend_slope'] == -1 and pv[0] > 0):
            pv = np.array([0, 0])

        self.pv = pv

        # detrend the data
        el_trend = self.mz[~self.nan_val] * pv[0] + pv[1]

        self.residuals = data[~self.nan_val] - el_trend

    def retrendData(self, r):
        """
        Retrend the residual values
        """

        # retrend the data
        return r + self.pv[0]*self.GridZ + self.pv[1]



import unittest
from io import StringIO
from unittest.mock import patch

from .script_test_helper import load_script_as_module

script = load_script_as_module('run_smrf')


class TestRunSmrf(unittest.TestCase):
    def setUp(self):
        self.argument_parser = script.argument_parser()

    @patch('sys.stderr', new=StringIO())
    def test_requires_config_file(self):
        with self.assertRaises(SystemExit):
            self.argument_parser.parse_args([])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
Input/Output functions
Adapted from the UW-Hydro tonic project
"""

from collections import Sequence

from .pycompat import basestring, unicode_type


def isbool(x):
    '''Test if str is an bolean'''
    if isinstance(x, float) or isinstance(x, basestring) and '.' in x:
        return False
    try:
        a = float(x)
        b = int(a)
    except ValueError:
        return False
    else:
        return a == b


def isfloat(x):
    '''Test if value is a float'''
    try:
        float(x)
    except ValueError:
        return False
    else:
        return True


def isint(x):
    '''Test if value is an integer'''
    if isinstance(x, float) or isinstance(x, basestring) and '.' in x:
        return False
    try:
        a = float(x)
        b = int(a)
    except ValueError:
        return False
    else:
        return a == b


def isscalar(x):
    '''Test if a value is a scalar'''
    if isinstance(x, (Sequence, basestring, unicode_type)):
        return False
    else:
        return True



import numpy as np


class IDW:
    '''
    Inverse distance weighting class for distributing input data. Availables
    options are:

    * Standard IDW
    * Detrended IDW

    '''

    def __init__(self, mx, my, GridX, GridY, mz=None, GridZ=None,
                 power=2, zeroVal=-1):
        """
        Args:
            mx: x locations for the points
            my: y locations for the points
            GridX: x locations in grid to interpolate over
            GridY: y locations in grid to interpolate over
            mz: elevation for the points
            GridZ: Elevation values for the points to interpolate over for
                   trended data
            power: power of the inverse distance weighting
        """

        # measurement point locations
        self.mx = mx
        self.my = my
        self.mz = mz
        self.npoints = len(mx)

        # grid information
        self.GridX = GridX
        self.GridY = GridY
        self.GridZ = GridZ

        # data information
        self.data = None
        self.nan_val = []

        # IDW parameters
        self.power = power
        self.zeroVal = zeroVal

        # calculate the distances
        self.calculateDistances()

        # calculate the weights
        self.calculateWeights()

    def calculateDistances(self):
        '''
        Calculate the distances from the measurement locations to the
        grid locations
        '''

        # preallocate
        self.distance = np.empty((self.GridX.shape[0],
                                  self.GridX.shape[1],
                                  self.npoints))

        for i in range(self.npoints):
            self.distance[:, :, i] = np.sqrt((self.GridX - self.mx[i])**2 +
                                             (self.GridY - self.my[i])**2)

        # remove any zero values
        self.distance[np.where(self.distance == 0)] = np.min(self.distance)

    def calculateWeights(self):
        '''
        Calculate the weights for
        '''

        # calculate the weights
        self.weights = 1.0/(np.power(self.distance, self.power))

        # if there are Inf values, set to 1 as the distance was 0
        # self.weights[np.isinf(self.weights)] = 100

    def calculateIDW(self, data, local=False):
        '''
        Calculate the IDW of the data at mx,my over GridX,GridY
        Inputs:
        data    - is the same size at mx,my
        '''
        nan_val = ~np.isnan(data)
        w = self.weights[:, :, nan_val]
        data = data[nan_val]

        v = np.nansum(w * data, 2) / np.sum(w, 2)

        return v

    def detrendedIDW(self, data, flag=0, zeros=None, local=False):
        '''
        Calculate the detrended IDW of the data at mx,my over GridX,GridY
        Inputs:
        data    - is the same size at mx,my
        '''

        self.detrendData(data, flag, zeros)
        v = self.calculateIDW(self.dtrend, local)
#         vtmp = v.copy()
        v = self.retrendData(v)

        if zeros is not None:
            v[v < 0] = 0
        return v

    def detrendData(self, data, flag=0, zeros=None):
        '''
        Detrend the data in val using the heights zmeas
        data    - is the same size at mx,my
        flag     - 1 for positive, -1 for negative, 0 for any trend imposed
        '''

        # calculate the trend on any real data
        nan_val = np.isnan(data)
        pv = np.polyfit(self.mz[~nan_val], data[~nan_val], 1)

        # apply trend constraints
        if flag == 1 and pv[0] < 0:
            pv = np.array([0, 0])
        elif (flag == -1 and pv[0] > 0):
            pv = np.array([0, 0])

        self.pv = pv

        # detrend the data
        el_trend = self.mz * pv[0] + pv[1]

        if zeros is not None:
            data[zeros] = self.zeroVal
        self.dtrend = data - el_trend

    def retrendData(self, idw):
        '''
        Retrend the IDW values
        '''

        # retrend the data
        return idw + self.pv[0]*self.GridZ + self.pv[1]



from smrf.distribute import image_data
from smrf.utils import utils


class ta(image_data.image_data):
    """
    The :mod:`~smrf.distribute.air_temp.ta` class allows for variable specific
    distributions that go beyond the base class.

    Air temperature is a relatively simple variable to distribute as it does
    not rely on any other variables, but has many variables that depend on it.
    Air temperature typically has a negative trend with elevation and performs
    best when detrended. However, even with a negative trend, it is possible to
    have instances where the trend does not apply, for example a temperature
    inversion or cold air pooling.  These types of conditions will have
    unintended consequences on variables that use the distributed air
    temperature.

    Args:
        taConfig: The [air_temp] section of the configuration file

    Attributes:
        config: configuration from [air_temp] section
        air_temp: numpy array of the air temperature
        stations: stations to be used in alphabetical order
    """

    variable = 'air_temp'

    # these are variables that can be output
    OUTPUT_VARIABLES = {
        'air_temp': {
            'units': 'degree_Celsius',
            'standard_name': 'air_temperature',
            'long_name': 'Air temperature'
        }
    }

    BASE_THREAD_VARIABLES = frozenset([
        'air_temp'
    ])

    # these are variables that are operate at the end only and do not need to
    # be written during main distribute loop
    post_process_variables = {}

    def __init__(self, taConfig):

        # extend the base class
        image_data.image_data.__init__(self, self.variable)

        # check and assign the configuration
        self.getConfig(taConfig)
        self._logger.debug('Created distribute.air_temp')

    def initialize(self, topo, data, date_time=None):
        """
        Initialize the distribution, solely calls
        :mod:`smrf.distribute.image_data.image_data._initialize`.

        Args:
            topo: :mod:`smrf.data.loadTopo.Topo` instance contain topographic
                data and infomation
            metadata: metadata Pandas dataframe containing the station metadata
                from :mod:`smrf.data.loadData` or :mod:`smrf.data.loadGrid`

        """

        self._logger.debug('Initializing distribute.air_temp')
        self.date_time = date_time
        self._initialize(topo, data.metadata)

    def distribute(self, data):
        """
        Distribute air temperature given a Panda's dataframe for a single time
        step. Calls :mod:`smrf.distribute.image_data.image_data._distribute`.

        Args:
            data: Pandas dataframe for a single time step from air_temp

        """

        self._logger.debug('{} Distributing air_temp'.format(data.name))

        self._distribute(data)
        self.air_temp = utils.set_min_max(self.air_temp, self.min, self.max)

    def distribute_thread(self, smrf_queue, data_queue):
        """
        Distribute the data using threading and queue. All data is provided
        and ``distribute_thread`` will go through each time step and call
        :mod:`smrf.distribute.air_temp.ta.distribute` then puts the distributed
        data into ``queue['air_temp']``.

        Args:
            queue: queue dictionary for all variables
            data: pandas dataframe for all data, indexed by date time

        """
        self._logger.info("Distributing {}".format(self.variable))

        for date_time in self.date_time:

            data = data_queue[self.variable].get(date_time)
            self.distribute(data)
            smrf_queue[self.variable].put([date_time, self.air_temp])



import numpy as np


class Susong1999:
    """
    Follows the IPW command mkprecip

    The precipitation phase, or the amount of precipitation falling as rain or
    snow, can significantly alter the energy and mass balance of the snowpack,
    either leading to snow accumulation or inducing melt :cite:`Marks&al:1998`
    :cite:`Kormos&al:2014`. The precipitation phase and initial snow density
    are based on the precipitation temperature (the distributed dew point
    temperature) and are estimated after Susong et al (1999)
    :cite:`Susong&al:1999`. The table below shows the relationship to
    precipitation temperature:

    ========= ======== ============ ===============
    Min Temp  Max Temp Percent snow Snow density
    [deg C]   [deg C]  [%]          [kg/m^3]
    ========= ======== ============ ===============
    -Inf      -5       100          75
    -5        -3       100          100
    -3        -1.5     100          150
    -1.5      -0.5     100          175
    -0.5      0        75           200
    0         0.5      25           250
    0.5       Inf      0            0
    ========= ======== ============ ===============
    """

    # Above table
    TABLE = [
        {
            'temp_min': -1e309,
            'temp_max': -5,
            'snow': 1,
            'density': 75
        },
        {
            'temp_min': -5,
            'temp_max': -3,
            'snow': 1,
            'density': 100
        },
        {
            'temp_min': -3,
            'temp_max': -1.5,
            'snow': 1,
            'density': 150
        },
        {
            'temp_min': -1.5,
            'temp_max': -0.5,
            'snow': 1,
            'density': 175
        },
        {
            'temp_min': -0.5,
            'temp_max': 0.0,
            'snow': 0.75,
            'density': 200
        },
        {
            'temp_min': 0.0,
            'temp_max': 0.5,
            'snow': 0.25,
            'density': 250
        },
        {
            'temp_min': 0.5,
            'temp_max': 1e309,
            'snow': 0,
            'density': 0
        }
    ]

    @staticmethod
    def run(temperature, precipitation):
        """
        Args:
            precipitation - numpy array of precipitation values [mm]

            temperature - array of temperature values, use dew point
                temperature if available [degrees C]

        Returns:
            dictionary:
           - **percent_snow** (*numpy.array*) - Percent of the precipitation
                that is snow in values 0.0-1.0.
           - **rho_s** (*numpy.array*) - Snow density values in kg/m^3.
        """
        ps = np.zeros(precipitation.shape)
        sd = np.zeros(ps.shape)

        # if no precipitation return all zeros
        if np.sum(precipitation) == 0:
            return {'pcs': ps, 'rho_s': sd}

        # Determine the indices and allocate based on the table above
        for row in Susong1999.TABLE:

            # Get values between the temperature ranges that have precipitation
            ind = ((temperature >= row['temp_min'])
                   & (temperature < row['temp_max']))
            # set the percent snow
            ps[ind] = row['snow']

            # set the density
            sd[ind] = row['density']

        # If there is no precipitation at a pixel, don't report a value
        # this may make isnobal crash, I'm not really sure
        ps[precipitation == 0] = 0
        sd[precipitation == 0] = 0

        return {'pcs': ps, 'rho_s': sd}



from smrf.framework.model_framework import SMRF
from smrf.tests.smrf_test_case_lakes import SMRFTestCaseLakes


class TestThreading(SMRFTestCaseLakes):
    """
    Test for all defined BASE_THREAD_VARIABLES, not accounting for config
    dependent additions.
    """

    def test_multi_thread_variables(self):
        config = self.thread_config()
        smrf = SMRF(config)
        smrf.loadTopo()
        smrf.create_distribution()

        self.assertCountEqual(
            smrf.distribute['air_temp'].thread_variables,
            ['air_temp']
        )

        self.assertCountEqual(
            smrf.distribute['vapor_pressure'].thread_variables,
            ['vapor_pressure', 'dew_point', 'precip_temp']
        )

        self.assertCountEqual(
            smrf.distribute['wind'].thread_variables,
            ['wind_speed', 'wind_direction']
        )

        self.assertCountEqual(
            smrf.distribute['precipitation'].thread_variables,
            [
                'precip',
                'percent_snow',
                'snow_density',
                'storm_days',
                'storm_total'
            ]
        )

        self.assertCountEqual(
            smrf.distribute['albedo'].thread_variables,
            ['albedo_vis', 'albedo_ir']
        )

        self.assertCountEqual(
            smrf.distribute['cloud_factor'].thread_variables,
            ['cloud_factor']
        )

        self.assertCountEqual(
            smrf.distribute['solar'].thread_variables,
            [
                'cloud_ir_beam', 'net_solar', 'cloud_vis_beam',
                'clear_ir_diffuse', 'veg_vis_diffuse', 'clear_ir_beam',
                'cloud_ir_diffuse', 'clear_vis_diffuse', 'veg_ir_diffuse',
                'veg_vis_beam', 'clear_vis_beam', 'veg_ir_beam',
                'cloud_vis_diffuse'
            ]
        )

        self.assertCountEqual(
            smrf.distribute['thermal'].thread_variables,
            ['thermal', 'thermal_clear']
        )

        self.assertCountEqual(
            smrf.distribute['soil_temp'].thread_variables,
            []
        )

        if hasattr(self, 'threads'):
            self.assertTrue(len(smrf.threads == 12))
